# -*- coding: utf-8 -*-
"""
datetime manager module.
"""

from datetime import datetime, date

import pytz

import pyrin.configuration.services as config_services
import pyrin.security.session.services as session_services
import pyrin.utils.datetime as datetime_utils

from pyrin.core.structs import Manager
from pyrin.globalization.datetime import DateTimePackage
from pyrin.globalization.datetime.enumerations import TimezoneEnum


class DateTimeManager(Manager):
    """
    datetime manager class.
    """

    package_class = DateTimePackage

    def __init__(self, **options):
        """
        initializes an instance of DateTimeManager.
        """

        super().__init__()

        server_timezone = config_services.get('globalization',
                                              'locale', 'babel_default_timezone')

        client_timezone = config_services.get('globalization', 'locale', 'client_timezone')

        self._server_timezone = self.get_timezone(server_timezone)
        self._client_timezone = self.get_timezone(client_timezone)

    def _try_add_timezone(self, value, timezone):
        """
        adds the given timezone info into input value if it has no timezone info.

        :param datetime value: value to add timezone info into it.
        :param str timezone: timezone name to get datetime based on it.

        :rtype: datetime
        """

        localized_value = value
        if value.tzinfo is None:
            localized_value = self._localize(value, timezone)

        return localized_value

    def _localize(self, value, timezone):
        """
        localizes input datetime with given timezone.

        input value should not have a timezone info.

        :param datetime value: value to be localized.
        :param str timezone: timezone name to get datetime based on it.

        :rtype: datetime
        """

        timezone = self.get_timezone(timezone)
        return timezone.localize(value)

    def _as_timezone(self, value, timezone):
        """
        gets the result of `astimezone` on the given value.

        :param datetime value: value to get normalized.
        :param str timezone: timezone name to get datetime based on it.

        :rtype: datetime
        """

        timezone = self.get_timezone(timezone)
        return value.astimezone(timezone)

    def _normalize(self, value, timezone):
        """
        normalizes input value using given timezone and returns it.

        :param datetime value: value to get normalized.
        :param str timezone: timezone name to get datetime based on it.

        :rtype: datetime
        """

        value = self._as_timezone(value, timezone)
        timezone = self.get_timezone(timezone)
        return timezone.normalize(value)

    def now(self, server=True, timezone=None):
        """
        gets the current datetime based on requested timezone.

        :param bool server: if set to True, server timezone will be used.
                            if set to False, client timezone will be used.
                            defaults to True.

        :param str timezone: timezone name to get datetime based on it.
                             if provided, the value of `server` input
                             will be ignored. defaults to None.

        :rtype: datetime
        """

        if timezone not in (None, ''):
            timezone = self.get_timezone(timezone)
        else:
            timezone = self.get_current_timezone(server=server)

        return datetime.now(timezone)

    def client_now(self):
        """
        gets the current datetime based on client timezone.

        this is a helper method to let get the client datetime
        without providing value to `now` method.

        :rtype: datetime
        """

        return self.now(server=False)

    def today(self, server=True, timezone=None):
        """
        gets the current date based on requested timezone.

        :param bool server: if set to True, server timezone will be used.
                            if set to False, client timezone will be used.
                            defaults to True.

        :param str timezone: timezone name to get date based on it.
                             if provided, the value of `server` input
                             will be ignored. defaults to None.

        :rtype: date
        """

        now = self.now(server, timezone=timezone)
        return date(year=now.year, month=now.month, day=now.day)

    def client_today(self):
        """
        gets the current date based on client timezone.

        this is a helper method to let get the client date
        without providing value to `today` method.

        :rtype: date
        """

        return self.today(server=False)

    def current_year(self, server=True, timezone=None):
        """
        gets the current year based on requested timezone.

        :param bool server: if set to True, server timezone will be used.
                            if set to False, client timezone will be used.
                            defaults to True.

        :param str timezone: timezone name to get year based on it.
                             if provided, the value of `server` input
                             will be ignored. defaults to None.

        :rtype: int
        """

        now = self.now(server=server, timezone=timezone)
        return now.year

    def current_client_year(self):
        """
        gets the current year based on client timezone.

        this is a helper method to let get the client year
        without providing value to `current_year` method.

        :rtype: int
        """

        client_now = self.client_now()
        return client_now.year

    def get_default_client_timezone(self):
        """
        gets the default client timezone.

        :rtype: tzinfo
        """

        return self._client_timezone

    def get_current_timezone(self, server):
        """
        gets the current timezone for server or client.

        :param bool server: if set to True, server timezone will be returned.
                            if set to False, client timezone will be returned.

        :rtype: tzinfo
        """

        if server is True:
            return self._server_timezone
        else:
            request = session_services.get_safe_current_request()
            timezone = None
            if request is not None:
                timezone = request.timezone

            return timezone or self._client_timezone

    def get_timezone(self, timezone):
        """
        gets the timezone based on given timezone name.

        :param str timezone: timezone name.

        :rtype: tzinfo
        """

        return pytz.timezone(timezone)

    def get_timezone_name(self, server):
        """
        gets the server or client timezone name.

        :param bool server: if set to True, server timezone name will be returned.
                            if set to False, client timezone name will be returned.

        :rtype: str
        """

        return self.get_current_timezone(server).zone

    def convert(self, value, to_server, from_server=None):
        """
        converts the given datetime between server and client timezones.

        :param datetime value: value to be converted.

        :param bool to_server: specifies that value must be normalized
                               to server timezone. if set to False, it
                               will be normalized to client timezone.

        :param bool from_server: specifies that value must be normalized
                                 from server timezone. if set to False, it
                                 will be normalized from client timezone.
                                 if not provided, it will be set to opposite
                                 of `to_server` value.
        :rtype: datetime
        """

        if from_server is None:
            from_server = not to_server

        localized_value = self.try_add_timezone(value, server=from_server)
        return self.normalize(localized_value, server=to_server)

    def convert_to_utc(self, value, from_server):
        """
        converts the given datetime to utc.

        :param datetime value: value to be converted.

        :param bool from_server: specifies that value must be normalized
                                 from server timezone. if set to False, it
                                 will be normalized from client timezone.

        :rtype: datetime
        """

        localized_value = self.try_add_timezone(value, server=from_server)
        return self._normalize(localized_value, TimezoneEnum.UTC)

    def convert_from_utc(self, value, to_server):
        """
        converts the given datetime from utc.

        :param datetime value: value to be converted.

        :param bool to_server: specifies that value must be normalized
                               to server timezone. if set to False, it
                               will be normalized to client timezone.

        :rtype: datetime
        """

        localized_value = self._try_add_timezone(value, TimezoneEnum.UTC)
        return self.normalize(localized_value, to_server)

    def as_timezone(self, value, server):
        """
        gets the result of `astimezone` on the given value.

        :param datetime value: value to get normalized.

        :param bool server: if set to True, server timezone name will be used.
                            if set to False, client timezone name will be used.

        :rtype: datetime
        """

        timezone = self.get_timezone_name(server=server)
        return self._as_timezone(value, timezone)

    def normalize(self, value, server):
        """
        normalizes input value using server or client current timezone and returns it.

        :param datetime value: value to get normalized.
        :param bool server: specifies that server or client timezone must used.

        :rtype: datetime
        """

        timezone = self.get_timezone_name(server)
        return self._normalize(value, timezone)

    def localize(self, value, server):
        """
        localizes input datetime with current timezone.

        input value should not have a timezone info.

        :param datetime value: value to be localized.
        :param bool server: specifies that server or client timezone must used.

        :rtype: datetime
        """

        timezone = self.get_timezone_name(server)
        return self._localize(value, timezone)

    def try_add_timezone(self, value, server):
        """
        adds the current timezone info into input value if it has no timezone info.

        :param datetime value: value to add timezone info into it.
        :param bool server: specifies that server or client timezone must be added.

        :rtype: datetime
        """

        timezone = self.get_timezone_name(server)
        return self._try_add_timezone(value, timezone)

    def to_datetime_string(self, value, to_server, from_server=None):
        """
        gets the datetime string representation of input value.

        if the value has no timezone info, it adds the client or server
        timezone info based on `from_server` value.

        example: `2015-12-24T22:40:15+01:00`

        :param datetime value: input object to be converted.

        :param bool to_server: specifies that value must be normalized
                               to server timezone. if set to False, it
                               will be normalized to client timezone.

        :param bool from_server: specifies that value must be normalized
                                 from server timezone. if set to False, it
                                 will be normalized from client timezone.
                                 if not provided, it will be set to opposite
                                 of `to_server` value.

        :rtype: str
        """

        localized_value = self.convert(value, to_server, from_server=from_server)
        return datetime_utils.to_datetime_string(localized_value)

    def to_date_string(self, value):
        """
        gets the date string representation of input value.

        example: `2015-12-24`

        :param datetime | date value: input object to be converted.

        :rtype: str
        """

        return datetime_utils.to_date_string(value)

    def to_time_string(self, value, to_server, from_server=None):
        """
        gets the time string representation of input value.

        if the value is a datetime and has no timezone info, it adds
        the client or server timezone info based on `from_server` value.

        example: `23:40:15`

        :param datetime | time value: input object to be converted.

        :param bool to_server: specifies that value must be normalized
                               to server timezone. if set to False, it
                               will be normalized to client timezone.

        :param bool from_server: specifies that value must be normalized
                                 from server timezone. if set to False, it
                                 will be normalized from client timezone.
                                 if not provided, it will be set to opposite
                                 of `to_server` value.

        :rtype: str
        """

        localized_value = value
        if isinstance(value, datetime):
            localized_value = self.convert(value, to_server, from_server=from_server)

        return datetime_utils.to_time_string(localized_value)

    def to_datetime(self, value, to_server, from_server=None):
        """
        converts the input value to it's equivalent python datetime.

        if the value has no timezone info, it adds the client or server
        timezone info based on `from_server` value.

        :param str value: string representation of datetime to be converted.

        :param bool to_server: specifies that value must be normalized
                               to server timezone. if set to False, it
                               will be normalized to client timezone.

        :param bool from_server: specifies that value must be normalized
                                 from server timezone. if set to False, it
                                 will be normalized from client timezone.
                                 if not provided, it will be set to opposite
                                 of `to_server` value.

        :rtype: datetime
        """

        converted_datetime = datetime_utils.to_datetime(value)
        return self.convert(converted_datetime, to_server, from_server=from_server)

    def to_date(self, value):
        """
        converts the input value to it's equivalent python date.

        :param str value: string representation of date to be converted.

        :rtype: date
        """

        return datetime_utils.to_date(value)

    def to_time(self, value):
        """
        converts the input value to it's equivalent python time.

        :param str value: string representation of time to be converted.

        :rtype: time
        """

        return datetime_utils.to_time(value)

    def timezone_exists(self, timezone_name):
        """
        gets a value indicating that a timezone with the given name exists.

        :param str timezone_name: timezone name to check for existence.

        :rtype: bool
        """

        return timezone_name in pytz.all_timezones_set

    def get_timestamp(self, value, date_sep='-', main_sep=' ',
                      time_sep=':', microsecond=False):
        """
        gets the timestamp with specified separators for given datetime.

        default format is `YYYY-MM-DD HH:mm:SS`.

        :param datetime value: datetime value to get its timestamp.

        :param str date_sep: a separator to put between date elements.
                             if set to None, no separator will be used.

        :param str main_sep: a separator to put between date and time part.
                             if set to None, no separator will be used.

        :param str time_sep: a separator to put between time elements.
                             if set to None, no separator will be used.

        :param bool microsecond: specifies that timestamp must include microseconds.
                                 defaults to False if not provided.

        :rtype: str
        """

        if date_sep is None:
            date_sep = ''
        if main_sep is None:
            main_sep = ''
        if time_sep is None:
            time_sep = ''

        place_holder = '{year}{date_sep}{month}{date_sep}{day}{main_sep}' \
                       '{hour}{time_sep}{minute}{time_sep}{second}'

        if microsecond is True:
            place_holder = place_holder + '.{microsecond}'

        return place_holder.format(year=str(value.year).zfill(4),
                                   month=str(value.month).zfill(2),
                                   day=str(value.day).zfill(2),
                                   hour=str(value.hour).zfill(2),
                                   minute=str(value.minute).zfill(2),
                                   second=str(value.second).zfill(2),
                                   date_sep=date_sep,
                                   main_sep=main_sep,
                                   time_sep=time_sep,
                                   microsecond=value.microsecond)

    def get_current_timestamp(self, date_sep='-', main_sep=' ',
                              time_sep=':', server=True,
                              timezone=None, microsecond=False):
        """
        gets the current timestamp with specified separators based on requested timezone.

        default format is `YYYY-MM-DD HH:mm:SS`.

        :param str date_sep: a separator to put between date elements.
                             if set to None, no separator will be used.

        :param str main_sep: a separator to put between date and time part.
                             if set to None, no separator will be used.

        :param str time_sep: a separator to put between time elements.
                             if set to None, no separator will be used.

        :param bool server: if set to True, server timezone will be used.
                            if set to False, client timezone will be used.
                            defaults to True.

        :param str timezone: timezone name to get datetime based on it.
                             if provided, the value of `server` input
                             will be ignored. defaults to None.

        :param bool microsecond: specifies that timestamp must include microseconds.
                                 defaults to False if not provided.

        :rtype: str
        """

        current = self.now(server=server, timezone=timezone)
        return self.get_timestamp(current, date_sep=date_sep, main_sep=main_sep,
                                  time_sep=time_sep, microsecond=microsecond)

    def datetime(self, year, month, day, hour=0, minute=0,
                 second=0, microsecond=0, fold=0,
                 server=True, timezone=None):
        """
        gets a new datetime with given inputs and requested timezone.

        :param int year: year.
        :param int month: month.
        :param int day: day.
        :param int hour: hour.
        :param int minute: minute.
        :param int second: second.
        :param int microsecond: microsecond.

        :param int fold: used to disambiguate wall times during a repeated
                         interval. it could be set to 0 or 1.

        :param bool server: if set to True, server timezone will be used.
                            if set to False, client timezone will be used.
                            defaults to True.

        :param str timezone: timezone name to be used.
                             if provided, the value of `server` input
                             will be ignored. defaults to None.
        :rtype: datetime
        """

        result = datetime(year=year, month=month, day=day, hour=hour, minute=minute,
                          second=second, microsecond=microsecond, fold=fold)

        if timezone not in (None, ''):
            return self._try_add_timezone(result, timezone)

        return self.try_add_timezone(result, server)



# -*- coding: utf-8 -*-
"""
template handlers exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class TemplateHandlersException(CoreException):
    """
    template handlers exception.
    """
    pass


class TemplateHandlersBusinessException(CoreBusinessException, TemplateHandlersException):
    """
    template handlers business exception.
    """
    pass


class InvalidPackagePathError(TemplateHandlersBusinessException):
    """
    invalid package path error.
    """
    pass


class InvalidPackageClassNameError(TemplateHandlersBusinessException):
    """
    invalid package class name error.
    """
    pass



# -*- coding: utf-8 -*-
"""
orm sql extractor manager module.
"""

import sqlparse

from sqlalchemy.sql.elements import TextClause
from sqlparse.tokens import DML, Keyword
from sqlparse.sql import IdentifierList, Identifier, Function, Where, Comparison, TokenList

import pyrin.utils.misc as misc_utils

from pyrin.core.structs import Manager
from pyrin.database.orm.sql.extractor import ORMSQLExtractorPackage


class ORMSQLExtractorManager(Manager):
    """
    orm sql extractor manager class.
    """

    package_class = ORMSQLExtractorPackage

    def find_table_names(self, expression, **options):
        """
        finds table names from a string or `TextClause` sql expression.

        :param str | TextClause expression: a string or `TextClause`
                                            containing a sql expression.

        :keyword bool include_select: specifies that select statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_insert: specifies that insert statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_update: specifies that update statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_delete: specifies that delete statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :rtype: list[str]
        """

        sql = expression
        tables = []
        if isinstance(expression, TextClause):
            sql = expression.text

        if sql is None or len(sql) <= 0:
            return tables

        return self._extract_tables(sql, **options)

    def _is_subselect(self, statement):
        """
        gets a value indicating that given statement is a select or sub-select.

        :param Statement statement: Statement object.

        :rtype: bool
        """

        if not statement.is_group:
            return False
        for item in statement.tokens:
            if item.ttype is DML and item.value.upper() == 'SELECT':
                return True
        return False

    def _is_insert(self, statement):
        """
        gets a value indicating that given statement is an insert.

        :param Statement statement: Statement object.

        :rtype: bool
        """

        if not statement.is_group:
            return False
        for item in statement.tokens:
            if item.ttype is DML and item.value.upper() == 'INSERT':
                return True
        return False

    def _is_update(self, statement):
        """
        gets a value indicating that given statement is an update.

        :param Statement statement: Statement object.

        :rtype: bool
        """

        if not statement.is_group:
            return False
        for item in statement.tokens:
            if item.ttype is DML and item.value.upper() == 'UPDATE':
                return True
        return False

    def _is_delete(self, statement):
        """
        gets a value indicating that given statement is a delete.

        :param Statement statement: Statement object.

        :rtype: bool
        """

        if not statement.is_group:
            return False
        for item in statement.tokens:
            if item.ttype is DML and item.value.upper() == 'DELETE':
                return True
        return False

    def _is_join(self, token):
        """
        gets a value indicating that given token represents a join keyword.

        :param Token token: token to be checked for join.

        :rtype: bool
        """

        if token.ttype is Keyword and (token.value.upper() == 'JOIN' or
                                       ' JOIN' in token.value.upper() or
                                       token.value.upper() == 'ON'):
            return True

        return False

    def _is_where(self, token):
        """
        gets a value indicating that given token represents a where clause.

        :param Token token: token to be checked for where clause.

        :rtype: bool
        """

        return isinstance(token, Where)

    def _is_comparison(self, token):
        """
        gets a value indicating that given token represents a comparison clause.

        :param Token token: token to be checked for comparison clause.

        :rtype: bool
        """

        return isinstance(token, Comparison)

    def _is_table(self, token):
        """
        gets a value indicating that given token represents a table.

        :param Token token: token to be checked for being table.

        :rtype: bool
        """

        return not self._is_join(token) and not self._is_where(token) and \
            not self._is_comparison(token)

    def _extract_select_from_part(self, statement):
        """
        extracts table names from a select statement.

        it also supports multiple joins and sub-selects.

        :param Statement statement: statement object.

        :rtype: list[Identifier]
        """

        select_seen = False
        from_seen = False
        join_seen = False
        results = []
        for item in statement.tokens:
            if item.is_group:
                for x in self._extract_select_from_part(item):
                    results.append(self._get_identifier(x, True))
            if select_seen is True and from_seen is True and join_seen is False:
                if self._is_subselect(item):
                    for z in self._extract_select_from_part(item):
                        results.append(self._get_identifier(z, True))
                elif item.ttype is Keyword and item.value.upper() in ['ORDER BY',
                                                                      'GROUP BY',
                                                                      'HAVING']:
                    return results
                elif self._is_join(item) is True:
                    join_seen = True
                elif item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    results.append(self._get_identifier(item, True))
            elif select_seen is True and from_seen is True and join_seen is True:
                if item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    results.append(self._get_identifier(item, True))
            elif item.ttype is Keyword.DML and item.value.upper() == 'SELECT':
                select_seen = True
            elif item.ttype is Keyword and item.value.upper() == 'FROM':
                from_seen = True
            elif self._is_join(item) is True:
                join_seen = True

        return results

    def _extract_insert_into_part(self, statement):
        """
        extracts table name from an insert statement.

        :param Statement statement: statement object.

        :rtype: Identifier
        """

        insert_seen = False
        into_seen = False
        for item in statement.tokens:
            if insert_seen is True and into_seen is True:
                if item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    return self._get_identifier(item, True)
            elif item.ttype is Keyword.DML and item.value.upper() == 'INSERT':
                insert_seen = True
            elif item.ttype is Keyword and item.value.upper() == 'INTO':
                into_seen = True

        return None

    def _extract_update_part(self, statement):
        """
        extracts table name from an update statement.

        :param Statement statement: statement object.

        :rtype: Identifier
        """

        update_seen = False
        for item in statement.tokens:
            if update_seen is True:
                if item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    return self._get_identifier(item, True)
            elif item.ttype is Keyword.DML and item.value.upper() == 'UPDATE':
                update_seen = True

        return None

    def _extract_delete_from_part(self, statement):
        """
        extracts table name from a delete statement.

        :param Statement statement: statement object.

        :rtype: Identifier
        """

        delete_seen = False
        from_seen = False
        for item in statement.tokens:
            if delete_seen is True and from_seen is True:
                if item.value not in (None, '') and not item.value.isspace() and \
                        self._is_table(item) is True:
                    return self._get_identifier(item, True)
            elif item.ttype is Keyword.DML and item.value.upper() == 'DELETE':
                delete_seen = True
            elif item.ttype is Keyword and item.value.upper() == 'FROM':
                from_seen = True

        return None

    def _extract_table_identifiers(self, token_stream):
        """
        extracts all potential table names from input single or collection `TokenList`.

        :param list[TokenList] | TokenList token_stream: a single `TokenList` or
                                                         collection of `TokenList`
                                                         objects to be checked.

        :rtype: list[str]
        """

        result = []
        if token_stream is not None:
            token_stream = misc_utils.make_iterable(token_stream, list)
            for item in token_stream:
                if item is not None:
                    if isinstance(item, IdentifierList):
                        for identifier in item.get_identifiers():
                            if isinstance(identifier, TokenList):
                                if identifier.get_real_name().lower() not in result:
                                    result.append(identifier.get_real_name().lower())
                            else:
                                if identifier.value.lower() not in result:
                                    result.append(identifier.value.lower())
                    elif isinstance(item, Identifier):
                        if item.get_real_name().lower() not in result:
                            result.append(item.get_real_name().lower())
                    elif item.ttype is Keyword:
                        if item.value not in result:
                            result.append(item.value.lower())
                    elif isinstance(item, Function):
                        if item.value not in result:
                            result.append(item.value.lower())

        return result

    def _extract_tables(self, sql, **options):
        """
        extracts all tables from given sql string expression.
        tables will be ordered as they appeared in input expression.

        :param str sql: sql expression be extracted for table names.

        :keyword bool include_select: specifies that select statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_insert: specifies that insert statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_update: specifies that update statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :keyword bool include_delete: specifies that delete statements
                                      must be investigated for table names.
                                      defaults to True if not provided.

        :rtype: list[str]
        """

        include_select = options.get('include_select', True)
        include_insert = options.get('include_insert', True)
        include_update = options.get('include_update', True)
        include_delete = options.get('include_delete', True)

        extracted_tables = {}
        statements = sqlparse.parse(sql)
        for statement in statements:
            stream = None
            if statement.get_type() != 'UNKNOWN':
                if include_select is True and self._is_subselect(statement):
                    stream = self._extract_select_from_part(statement)
                    extracted_tables.update(**dict.fromkeys(
                        self._extract_table_identifiers(stream), None))
                if include_insert is True and self._is_insert(statement):
                    stream = self._extract_insert_into_part(statement)
                    extracted_tables.update(**dict.fromkeys(
                        self._extract_table_identifiers(stream), None))
                if include_update is True and self._is_update(statement):
                    stream = self._extract_update_part(statement)
                    extracted_tables.update(**dict.fromkeys(
                        self._extract_table_identifiers(stream), None))
                if include_delete is True and self._is_delete(statement):
                    stream = self._extract_delete_from_part(statement)
                    extracted_tables.update(**dict.fromkeys(
                        self._extract_table_identifiers(stream), None))

        return list(extracted_tables.keys())

    def _get_identifier(self, token, force=False):
        """
        gets the identifier token available in given tokens children.

        :param Token | TokenList token: token to get its child identifier.

        :param bool force: specifies that if its child is not an
                           `Identifier` but it is a keyword, return it.
                           it could be helpful for some situations which
                           table names are equal to some keywords.

        :rtype: Identifier | Keyword
        """

        if token is not None:
            if isinstance(token, (Identifier, IdentifierList)):
                return token

            elif isinstance(token, Function) and force is True:
                return token.tokens[0]

            elif token.ttype is Keyword and force is True:
                return token

        return None



# -*- coding: utf-8 -*-
"""
token decorators module.
"""

import pyrin.security.token.services as token_services


def token(*args, **kwargs):
    """
    decorator to register a token handler.

    :param object args: token handler class constructor arguments.
    :param object kwargs: token handler class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           handler with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidTokenHandlerTypeError: invalid token handler type error.
    :raises InvalidTokenHandlerNameError: invalid token handler name error.
    :raises DuplicatedTokenHandlerError: duplicated token handler error.
    :raises DuplicatedTokenKidHeaderError: duplicated token kid header error.

    :returns: token handler class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available token handlers.

        :param type cls: token handler class.

        :returns: token handler class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        token_services.register_token_handler(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
database handlers module.
"""

import pyrin.application.services as application_services
import pyrin.database.services as database_services


application_services.register_teardown_request_handler(database_services.cleanup_session)



# -*- coding: utf-8 -*-
"""
orm scoping exceptions module.
"""

from sqlalchemy.exc import InvalidRequestError

from pyrin.core.exceptions import CoreException


class ORMScopingException(CoreException):
    """
    orm scoping exception.
    """
    pass


class InvalidRequestScopedRegistryTypeError(ORMScopingException):
    """
    invalid request scoped registry type error.
    """
    pass


class InvalidThreadScopedRegistryTypeError(ORMScopingException):
    """
    invalid thread scoped registry type error.
    """
    pass


class ScopedSessionIsAlreadyPresentError(InvalidRequestError, ORMScopingException):
    """
    scoped session is already present error.
    """
    pass



# -*- coding: utf-8 -*-
"""
string normalizer handlers base module.
"""

import re

from abc import abstractmethod
from collections import OrderedDict

import pyrin.utils.dictionary as dict_utils

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.utilities.string.normalizer.interface import AbstractStringNormalizerBase
from pyrin.utilities.string.normalizer.handlers.exceptions import FilterMapMustBeDictError, \
    InvalidStringNormalizerNameError, InvalidStringNormalizerPriorityError, \
    FiltersMustBeListError


class StringNormalizerBase(AbstractStringNormalizerBase):
    """
    string normalizer base class.
    """

    def __init__(self, name, priority, *args, **options):
        """
        initializes an instance of StringNormalizerBase.

        :param str name: name of this normalizer.
                         the normalizer will be registered by this name
                         into available normalizers. it must be unique.

        :param int priority: priority of this normalizer.
                             normalizers with higher priority will be executed sooner.

        :raises InvalidStringNormalizerNameError: invalid string normalizer name error.
        :raises InvalidStringNormalizerPriorityError: invalid string normalizer priority error.
        """

        super().__init__()

        if name in (None, '') or name.isspace():
            raise InvalidStringNormalizerNameError('String normalizer name must be '
                                                   'provided for [{normalizer}].'
                                                   .format(normalizer=self))

        if not isinstance(priority, int):
            raise InvalidStringNormalizerPriorityError('String normalizer [{normalizer}] '
                                                       'priority must be an integer.'
                                                       .format(normalizer=name))

        self._set_name(name)
        self._priority = priority

    def normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :keyword bool strip: strip spaces from both ends of value.
                             defaults to True if not provided.

        :keyword bool normalize_none: specifies that if given value is None,
                                      return empty string instead of None.
                                      defaults to False if not provided.

        :returns: normalized value.
        :rtype: str
        """

        normalize_none = options.get('normalize_none', False)
        if value is None and normalize_none is True:
            value = ''

        if value in (None, ''):
            return value

        value = self._normalize(value, **options)
        strip = options.get('strip', True)
        if strip is not False:
            value = value.strip()

        return value

    @abstractmethod
    def _normalize(self, value, **options):
        """
        normalizes the given value.

        this method is intended to be overridden by subclasses.

        :param str value: value to be normalized.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: normalized value.
        :rtype: str
        """

        raise CoreNotImplementedError()

    @property
    def priority(self):
        """
        gets the priority of this normalizer.

        :rtype: int
        """

        return self._priority


class FilterNormalizerBase(StringNormalizerBase):
    """
    filter normalizer base class.

    this normalizer will filter provided values from string.
    """

    def __init__(self, name, priority, **options):
        """
        initializes an instance of FilterNormalizerBase.

        :param str name: name of this normalizer.
                         the normalizer will be registered by this name
                         into available normalizers. it must be unique.

        :param int priority: priority of this normalizer.
                             normalizers with higher priority will be executed sooner.

        :keyword list[str] filters: values to be removed from string.

        :keyword dict filter_map: a dict of keys and values to be used
                                  for replacement.

        :note filters, filter_map: the `filters` value have precedence
                                   over `filter_map` value if they have
                                   common keys.

        :raises InvalidStringNormalizerNameError: invalid string normalizer name error.
        :raises InvalidStringNormalizerPriorityError: invalid string normalizer priority error.
        :raises FiltersMustBeListError: filters must be list error.
        :raises FilterMapMustBeDictError: filter map must be dict error.
        """

        super().__init__(name, priority, **options)

        filters = options.get('filters')
        filter_map = options.get('filter_map')

        if filters is not None and not isinstance(filters, list):
            raise FiltersMustBeListError('The provided value for "filters" must be a list.')

        if filter_map is not None and not isinstance(filter_map, dict):
            raise FilterMapMustBeDictError('The provided value for "filter_map" must be a dict.')

        self._filter_map = self._join(filters, filter_map)

    def _join(self, filters, filter_map):
        """
        joins given `filters` and `filter_map` into a single ordered dict.

        the keys will be sorted by length.

        :param list[str] filters: values to be removed from string.

        :param dict filter_map: a dict of keys and values to be used
                                for replacement.

        :note filters, filter_map: the `filters` value have precedence
                                   over `filter_map` value if they have
                                   common keys.

        :rtype: OrderedDict
        """

        if filters is None and filter_map is None:
            return OrderedDict()

        result = dict()
        if filters is not None:
            filters = set(filters)
            for item in filters:
                result[item] = ''

        if filter_map is not None:
            filter_map.update(result)
            result = filter_map

        result = dict_utils.sort_by_key_length(result, reverse=True)
        return result

    def _combine_with_current(self, filters, filter_map):
        """
        combines given inputs with current `filter_maps` and returns a new dict.

        :param list[str] filters: values to be removed from string.

        :param dict filter_map: a dict of keys and values to be used
                                for replacement.

        :note filters, filter_map: the `filters` value have precedence
                                   over `filter_map` value if they have
                                   common keys.

        :rtype: OrderedDict
        """

        if filter_map is None:
            filter_map = dict()

        current = dict(self._filter_map)
        current.update(filter_map)
        filter_map = current
        return self._join(filters, filter_map)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :keyword list[str] filters: values to be removed from string.
                                    this value will be combined with
                                    `self._filter_map` if provided.

        :keyword dict filter_map: a dict of keys and values to be used
                                  for replacement. this value will be
                                  combined with `self._filter_map` if
                                  provided.

        :note filters, filter_map: the `filters` value have precedence
                                   over `filter_map` value if they have
                                   common keys.

        :keyword bool ignore_case: filter matches in case-insensitive way.
                                   defaults to True if not provided.

        :returns: normalized value.
        :rtype: str
        """

        flags = re.IGNORECASE
        ignore_case = options.get('ignore_case', True)
        if ignore_case is False:
            flags = 0

        filter_map = self._combine_with_current(options.get('filters'),
                                                options.get('filter_map'))
        for key, item in filter_map.items():
            value = re.sub(key, item, value, flags=flags)

        return value


class ReplaceNormalizerBase(StringNormalizerBase):
    """
    replace normalizer class.

    this normalizer replaces or removes provided values from string.
    """

    def __init__(self, name, priority, replace_map, **options):
        """
        initializes an instance of ReplaceNormalizerBase.

        :param str name: name of this normalizer.
                         the normalizer will be registered by this name
                         into available normalizers. it must be unique.

        :param int priority: priority of this normalizer.
                             normalizers with higher priority will be executed sooner.

        :param dict replace_map: a dict of keys and values to
                                 be used for replacement or removal.
                                 note that keys must be single length characters.
                                 to remove a key from string, set the key with None
                                 value in the `replace_map`.

        :raises InvalidStringNormalizerNameError: invalid string normalizer name error.
        :raises InvalidStringNormalizerPriorityError: invalid string normalizer priority error.
        """

        super().__init__(name, priority, **options)

        self._map = replace_map
        self._translation_table = str.maketrans(replace_map)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :returns: normalized value.
        :rtype: str
        """

        return value.translate(self._translation_table)



# -*- coding: utf-8 -*-
"""
model mixin module.

this module provides mixin classes for different features of declarative base entity.
if you want to implement a new declarative base class and not use the `CoreEntity`
provided by pyrin, you could define your new base class and it must be inherited
from `BaseEntity`, because application will check isinstance() on `BaseEntity` type
to detect models. and then implement your customized or new features in your base class.
then you must put `metaclass=DeclarativeMeta` on your new base class. now all your concrete
entities must be inherited from the new declarative base class. note that you must use a
unique declarative base class for all your models, do not mix `CoreEntity` and your new
declarative base class usage. otherwise you will face problems in migrations and also
multi-database environments.
"""

import inspect

from abc import abstractmethod

from sqlalchemy import inspect as sqla_inspect, UniqueConstraint
from sqlalchemy.orm import declared_attr
from sqlalchemy.exc import NoInspectionAvailable
from sqlalchemy.ext.hybrid import hybrid_property

import pyrin.globalization.datetime.services as datetime_services
import pyrin.database.model.services as model_services
import pyrin.configuration.services as config_services
import pyrin.utils.sqlalchemy as sqlalchemy_utils
import pyrin.utils.misc as misc_utils

from pyrin.core.globals import _
from pyrin.core.decorators import class_property
from pyrin.caching.mixin.decorators import fast_cache
from pyrin.caching.mixin.typed import TypedCacheMixin
from pyrin.core.globals import LIST_TYPES, SECURE_TRUE, SECURE_FALSE
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.orm.sql.schema.columns import TimeStampColumn
from pyrin.utils.custom_print import print_warning
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.database.services import get_current_store
from pyrin.core.structs import DTO, CoreImmutableDict
from pyrin.utils.exceptions import InvalidOrderingColumnError
from pyrin.database.model.exceptions import ColumnNotExistedError, \
    InvalidDeclarativeBaseTypeError, InvalidDepthProvidedError, InvalidOrderingColumnTypeError


class ModelMixinBase:
    """
    model mixin base class.

    all model mixins that have cached values must be subclassed from this.
    """

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.

        this method will be get called on server startup for each entity.
        subclasses must call `super().populate_cache()` at the end.
        """
        pass


class ColumnMixin(ModelMixinBase):
    """
    column mixin class.

    this class adds functionalities about columns (other than pk and fk) to its subclasses.
    """

    @class_property
    @fast_cache
    def all_columns(cls):
        """
        gets all column names of this entity.

        note that primary and foreign keys are not included in columns.
        column names will be calculated once and cached.

        note:
        as a column is either readable or not readable, so readable columns and
        not readable columns equal to all columns, as is for writable and
        not writable columns.

        :rtype: tuple[str]
        """

        return cls.readable_columns + cls.not_readable_columns

    @class_property
    @fast_cache
    def readable_columns(cls):
        """
        gets readable column names of this entity.

        which are those that have `allow_read=True` in their definition
        and their name does not start with underscore `_`.
        note that primary and foreign keys are not included in columns.
        column names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        columns = tuple(attr.key for attr in info.column_attrs
                        if cls.is_public(attr.key) is True and
                        attr.columns[0].is_foreign_key is False and
                        attr.columns[0].primary_key is False and
                        attr.columns[0].allow_read is True)

        return columns

    @class_property
    @fast_cache
    def not_readable_columns(cls):
        """
        gets not readable column names of this entity.

        which are those that have `allow_read=False` in their definition
        or their name starts with underscore `_`.
        note that primary and foreign keys are not included in columns.
        column names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        columns = tuple(attr.key for attr in info.column_attrs
                        if (cls.is_public(attr.key) is False or
                            attr.columns[0].allow_read is False) and
                        attr.columns[0].is_foreign_key is False and
                        attr.columns[0].primary_key is False)

        return columns

    @class_property
    @fast_cache
    def writable_columns(cls):
        """
        gets writable column names of this entity.

        which are those that have `allow_write=True` in their definition
        and their name does not start with underscore `_`.
        note that primary and foreign keys are not included in columns.
        column names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        columns = tuple(attr.key for attr in info.column_attrs
                        if cls.is_public(attr.key) is True and
                        attr.columns[0].is_foreign_key is False and
                        attr.columns[0].primary_key is False and
                        attr.columns[0].allow_write is True)

        return columns

    @class_property
    @fast_cache
    def not_writable_columns(cls):
        """
        gets not writable column names of this entity.

        which are those that have `allow_write=False` in their definition
        or their name starts with underscore `_`.
        note that primary and foreign keys are not included in columns.
        column names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        columns = tuple(attr.key for attr in info.column_attrs
                        if (cls.is_public(attr.key) is False or
                            attr.columns[0].allow_write is False) and
                        attr.columns[0].is_foreign_key is False and
                        attr.columns[0].primary_key is False)

        return columns

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.readable_columns
        temp = cls.not_readable_columns
        temp = cls.writable_columns
        temp = cls.not_writable_columns
        super().populate_cache()


class RelationshipMixin(ModelMixinBase):
    """
    relationship mixin class.

    this class adds functionalities about relationship properties to its subclasses.
    """

    @class_property
    @fast_cache
    def relationships(cls):
        """
        gets all relationship property names of this entity.

        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        return cls.exposed_relationships + cls.not_exposed_relationships

    @class_property
    @fast_cache
    def exposed_relationships(cls):
        """
        gets exposed relationship property names of this entity.

        which are those that their name does not start with underscore `_`.
        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        relationships = tuple(attr.key for attr in info.relationships
                              if cls.is_public(attr.key) is True)
        return relationships

    @class_property
    @fast_cache
    def not_exposed_relationships(cls):
        """
        gets not exposed relationship property names of this entity.

        which are those that their name starts with underscore `_`.
        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        relationships = tuple(attr.key for attr in info.relationships
                              if cls.is_public(attr.key) is False)
        return relationships

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.relationships
        super().populate_cache()


class HybridPropertyMixin(ModelMixinBase):
    """
    hybrid property mixin class.

    this class adds functionalities about all hybrid properties to its subclasses.
    """

    @class_property
    @fast_cache
    def all_getter_hybrid_properties(cls):
        """
        gets all getter hybrid property names of this entity.

        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        return cls.readable_hybrid_properties + cls.not_readable_hybrid_properties

    @class_property
    @fast_cache
    def all_setter_hybrid_properties(cls):
        """
        gets all setter hybrid property names of this entity.

        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        return cls.writable_hybrid_properties + cls.not_writable_hybrid_properties

    @class_property
    @fast_cache
    def readable_hybrid_properties(cls):
        """
        gets readable hybrid property names of this entity.

        readable hybrid properties are those that their name does
        not start with underscore `_`.
        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        hybrid_properties = tuple(item.__name__ for item in info.all_orm_descriptors
                                  if isinstance(item, hybrid_property)
                                  and item.fget is not None
                                  and cls.is_public(item.__name__) is True)

        return hybrid_properties

    @class_property
    @fast_cache
    def not_readable_hybrid_properties(cls):
        """
        gets not readable hybrid property names of this entity.

        not readable hybrid properties are those that their
        name starts with underscore `_`.
        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        hybrid_properties = tuple(item.__name__ for item in info.all_orm_descriptors
                                  if isinstance(item, hybrid_property)
                                  and item.fget is not None
                                  and cls.is_public(item.__name__) is False)

        return hybrid_properties

    @class_property
    @fast_cache
    def writable_hybrid_properties(cls):
        """
        gets writable hybrid property names of this entity.

        writable hybrid properties are those that their name does
        not start with underscore `_`.
        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        hybrid_properties = tuple(item.__name__ for item in info.all_orm_descriptors
                                  if isinstance(item, hybrid_property)
                                  and item.fset is not None
                                  and cls.is_public(item.__name__) is True)

        return hybrid_properties

    @class_property
    @fast_cache
    def not_writable_hybrid_properties(cls):
        """
        gets not writable hybrid property names of this entity.

        not writable hybrid properties are those that their
        name starts with underscore `_`.
        property names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        hybrid_properties = tuple(item.__name__ for item in info.all_orm_descriptors
                                  if isinstance(item, hybrid_property)
                                  and item.fset is not None
                                  and cls.is_public(item.__name__) is False)

        return hybrid_properties

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.all_getter_hybrid_properties
        temp = cls.all_setter_hybrid_properties
        super().populate_cache()


class PrimaryKeyMixin(ModelMixinBase):
    """
    primary key mixin class.

    this class adds functionalities about primary keys to its subclasses.
    """

    def _is_primary_key_comparable(self, primary_key):
        """
        gets a value indicating that given primary key is comparable.

        the primary key is comparable if it is not None for single
        primary keys and if all the values in primary key tuple are
        not None for composite primary keys.

        :param object | tuple[object] primary_key: primary key value.

        :rtype: bool
        """

        if primary_key is None:
            return False

        if isinstance(primary_key, tuple):
            if len(primary_key) <= 0:
                return False
            else:
                return all(pk is not None for pk in primary_key)

        return True

    def primary_key(self, as_tuple=False):
        """
        gets the primary key value for this entity.

        it could be a single value or a tuple of values
        for composite primary keys.
        it could return None if no primary key is set for this entity.

        :param bool as_tuple: specifies that primary key value must be returned
                              as a tuple even if it's a single value.
                              defaults to False if not provided.

        :rtype: object | tuple[object]
        """

        columns = self.primary_key_columns
        if len(columns) <= 0:
            return None

        if as_tuple is False and len(columns) == 1:
            return getattr(self, columns[0])
        else:
            return tuple(getattr(self, col) for col in columns)

    @class_property
    @fast_cache
    def primary_key_columns(cls):
        """
        gets all primary key column names of this entity.

        column names will be calculated once and cached.

        note:
        as a column is either readable or not readable, so readable columns and
        not readable columns equal to all columns, as is for writable and
        not writable columns.

        :rtype: tuple[str]
        """

        return cls.readable_primary_key_columns + cls.not_readable_primary_key_columns

    @class_property
    @fast_cache
    def readable_primary_key_columns(cls):
        """
        gets the readable primary key column names of this entity.

        which are those that have `allow_read=True` in their definition
        and their name does not start with underscore `_`.
        column names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        pk = tuple(info.get_property_by_column(col).key for col in info.primary_key
                   if cls.is_public(info.get_property_by_column(col).key) is True
                   and col.allow_read is True)

        return pk

    @class_property
    @fast_cache
    def not_readable_primary_key_columns(cls):
        """
        gets not readable primary key column names of this entity.

        which are those that have `allow_read=False` in their definition
        or their name starts with underscore `_`.
        column names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        pk = tuple(info.get_property_by_column(col).key for col in info.primary_key
                   if cls.is_public(info.get_property_by_column(col).key) is False
                   or col.allow_read is False)

        return pk

    @class_property
    @fast_cache
    def writable_primary_key_columns(cls):
        """
        gets the writable primary key column names of this entity.

        which are those that have `allow_write=True` in their definition
        and their name does not start with underscore `_`.
        column names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        pk = tuple(info.get_property_by_column(col).key for col in info.primary_key
                   if cls.is_public(info.get_property_by_column(col).key) is True
                   and col.allow_write is True)

        return pk

    @class_property
    @fast_cache
    def not_writable_primary_key_columns(cls):
        """
        gets not writable primary key column names of this entity.

        which are those that have `allow_write=False` in their definition
        or their name starts with underscore `_`.
        column names will be calculated once and cached.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        pk = tuple(info.get_property_by_column(col).key for col in info.primary_key
                   if cls.is_public(info.get_property_by_column(col).key) is False
                   or col.allow_write is False)

        return pk

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.readable_primary_key_columns
        temp = cls.not_readable_primary_key_columns
        temp = cls.writable_primary_key_columns
        temp = cls.not_writable_primary_key_columns
        super().populate_cache()


class ForeignKeyMixin(ModelMixinBase):
    """
    foreign key mixin class.

    this class adds functionalities about foreign keys to its subclasses.
    """

    @class_property
    @fast_cache
    def foreign_key_columns(cls):
        """
        gets all foreign key column names of this entity.

        column names will be calculated once and cached.
        note that those foreign keys which are also a primary key, will not
        be included in this list.

        note:
        as a column is either readable or not readable, so readable columns and
        not readable columns equal to all columns, as is for writable and
        not writable columns.

        :rtype: tuple[str]
        """

        return cls.readable_foreign_key_columns + cls.not_readable_foreign_key_columns

    @class_property
    @fast_cache
    def readable_foreign_key_columns(cls):
        """
        gets the readable foreign key column names of this entity.

        which are those that have `allow_read=True` in their definition
        and their name does not start with underscore `_`.
        column names will be calculated once and cached.
        note that those foreign keys which are also a primary key, will not
        be included in this list.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        fk = tuple(attr.key for attr in info.column_attrs
                   if attr.columns[0].is_foreign_key is True
                   and attr.columns[0].primary_key is False
                   and cls.is_public(attr.key) is True
                   and attr.columns[0].allow_read is True)

        return fk

    @class_property
    @fast_cache
    def not_readable_foreign_key_columns(cls):
        """
        gets not readable foreign key column names of this entity.

        which are those that have `allow_read=False` in their definition
        or their name starts with underscore `_`.
        column names will be calculated once and cached.
        note that those foreign keys which are also a primary key, will not
        be included in this list.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        fk = tuple(attr.key for attr in info.column_attrs
                   if attr.columns[0].is_foreign_key is True
                   and attr.columns[0].primary_key is False and
                   (cls.is_public(attr.key) is False or
                    attr.columns[0].allow_read is False))

        return fk

    @class_property
    @fast_cache
    def writable_foreign_key_columns(cls):
        """
        gets the writable foreign key column names of this entity.

        which are those that have `allow_write=True` in their definition
        and their name does not start with underscore `_`.
        column names will be calculated once and cached.
        note that those foreign keys which are also a primary key, will not
        be included in this list.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        fk = tuple(attr.key for attr in info.column_attrs
                   if attr.columns[0].is_foreign_key is True
                   and attr.columns[0].primary_key is False
                   and cls.is_public(attr.key) is True
                   and attr.columns[0].allow_write is True)

        return fk

    @class_property
    @fast_cache
    def not_writable_foreign_key_columns(cls):
        """
        gets not writable foreign key column names of this entity.

        which are those that have `allow_write=False` in their definition
        or their name starts with underscore `_`.
        column names will be calculated once and cached.
        note that those foreign keys which are also a primary key, will not
        be included in this list.

        :rtype: tuple[str]
        """

        info = sqla_inspect(cls)
        fk = tuple(attr.key for attr in info.column_attrs
                   if attr.columns[0].is_foreign_key is True
                   and attr.columns[0].primary_key is False and
                   (cls.is_public(attr.key) is False or
                    attr.columns[0].allow_write is False))

        return fk

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.readable_foreign_key_columns
        temp = cls.not_readable_foreign_key_columns
        temp = cls.writable_foreign_key_columns
        temp = cls.not_writable_foreign_key_columns
        super().populate_cache()


class AttributeMixin(ModelMixinBase):
    """
    attribute mixin class.

    this class adds functionalities about all attributes to its subclasses.
    attributes includes pk, fk, columns, relationships and hybrid properties.
    """

    @class_property
    @fast_cache
    def all_column_attributes(cls):
        """
        gets an immutable dict of all column attributes of this entity.

        the result will be calculated once and cached per entity type.

        :returns: CoreImmutableDict(str name, CoreColumn column)
        :rtype: CoreImmutableDict
        """

        result = dict()
        info = sqla_inspect(cls)
        for attr in info.column_attrs:
            result[attr.key] = attr.columns[0]

        return CoreImmutableDict(result)

    @class_property
    @fast_cache
    def all_reverse_column_attributes(cls):
        """
        gets an immutable dict of all reverse column attributes of this entity.

        this is the opposite of `all_column_attributes` value. the result
        dict keys are column instances and their values are attribute names.
        the result will be calculated once and cached per entity type.

        :returns: CoreImmutableDict(CoreColumn column, str name)
        :rtype: CoreImmutableDict
        """

        result = dict()
        for name, column in cls.all_column_attributes.items():
            result[column] = name

        return CoreImmutableDict(result)

    @class_property
    @fast_cache
    def all_instrumented_attributes(cls):
        """
        gets a tuple of all instrumented attributes of this entity.

        instrumented attributes are the attributes that are returned
        by sqlalchemy when you get a column of a model at class level.

        for example:

        CarEntity.name -> InstrumentedAttribute
        HumanEntity.age -> InstrumentedAttribute

        the result will be calculated once and cached per entity type.

        :rtype: tuple
        """

        all_columns = cls.primary_key_columns + cls.foreign_key_columns + cls.all_columns
        result = []
        for name in all_columns:
            attribute = getattr(cls, name)
            result.append(attribute)

        return tuple(result)

    @class_property
    @fast_cache
    def all_attributes(cls):
        """
        gets all attribute names of current entity.

        attribute names will be calculated once and cached.

        :rtype: tuple[str]
        """

        return cls.all_readable_attributes + cls.all_not_readable_attributes

    @class_property
    @fast_cache
    def all_readable_attributes(cls):
        """
        gets all readable attribute names of current entity.

        which are those that have `allow_read=True` (only for columns) in
        their definition and their name does not start with underscore `_`.
        attribute names will be calculated once and cached.

        :rtype: tuple[str]
        """

        return cls.readable_primary_key_columns + cls.readable_foreign_key_columns + \
            cls.readable_columns + cls.exposed_relationships + cls.readable_hybrid_properties

    @class_property
    @fast_cache
    def all_not_readable_attributes(cls):
        """
        gets all not readable attribute names of current entity.

        which are those that have `allow_read=False` (only for columns) in
        their definition or their name starts with underscore `_`.
        attribute names will be calculated once and cached.

        :rtype: tuple[str]
        """

        return cls.not_readable_primary_key_columns + cls.not_readable_foreign_key_columns + \
            cls.not_readable_columns + cls.not_exposed_relationships + \
            cls.not_readable_hybrid_properties

    @class_property
    @fast_cache
    def all_writable_attributes(cls):
        """
        gets all writable attribute names of current entity.

        which are those that have `allow_write=True` (only for columns) in
        their definition and their name does not start with underscore `_`.
        attribute names will be calculated once and cached.

        :rtype: tuple[str]
        """

        return cls.writable_primary_key_columns + cls.writable_foreign_key_columns + \
            cls.writable_columns + cls.exposed_relationships + cls.writable_hybrid_properties

    @class_property
    @fast_cache
    def all_not_writable_attributes(cls):
        """
        gets all not writable attribute names of current entity.

        which are those that have `allow_write=False` (only for columns) in
        their definition or their name starts with underscore `_`.
        attribute names will be calculated once and cached.

        :rtype: tuple[str]
        """

        return cls.not_writable_primary_key_columns + cls.not_writable_foreign_key_columns + \
            cls.not_writable_columns + cls.not_exposed_relationships + \
            cls.not_writable_hybrid_properties

    @classmethod
    def is_public(cls, name):
        """
        gets a value indicating that an attribute with given name is public.

        it simply checks that the given name starts with an underscore `_`.
        if so, it is considered as not public.

        :param str name: attribute name.

        :rtype: bool
        """

        return not name.startswith('_')

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.all_column_attributes
        temp = cls.all_reverse_column_attributes
        temp = cls.all_instrumented_attributes
        temp = cls.all_readable_attributes
        temp = cls.all_not_readable_attributes
        temp = cls.all_writable_attributes
        temp = cls.all_not_writable_attributes
        super().populate_cache()


class ConverterMixin:
    """
    converter mixin class.

    this class adds functionalities to convert dict to
    entity and vice versa to its subclasses.
    """

    # maximum allowed depth for conversion.
    # note that higher depth values may cause performance issues or
    # application failure in some cases. so if you do not know how
    # much depth is required for conversion, start without providing depth.
    # this value could be overridden in concrete entities if required.
    MAX_DEPTH = 5

    def to_dict(self, **options):
        """
        converts the entity into a dict and returns it.

        it could convert primary keys, foreign keys, other columns, hybrid
        properties and also relationship properties if `depth` is provided.
        the result dict by default only contains the readable attributes of the
        entity which are those that have `allow_read=True` (only for columns) and
        their name does not start with underscore `_`.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided.

        :keyword dict[str, list[str]] | list[str] columns: column names to be included in result.
                                                           it could be a list of column names.
                                                           for example:
                                                           `columns=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity` and
                                                           `PersonEntity`, it should be like this:
                                                           `columns=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                           if provided column names are not
                                                           available in result, they will
                                                           be ignored.

        :note columns: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword dict[str, dict[str, str]] | dict[str, str] rename: column names that must be
                                                                    renamed in the result.
                                                                    it could be a dict with keys
                                                                    as original column names and
                                                                    values as new column names
                                                                    that should be exposed instead
                                                                    of original column names.
                                                                    for example:
                                                                    `rename=dict(age='new_age',
                                                                                 name='new_name')`
                                                                    but if you want to include
                                                                    relationships, then you must
                                                                    provide a dict containing
                                                                    entity class name as key and
                                                                    for value, another dict
                                                                    containing original column
                                                                    names as keys, and column
                                                                    names that must be exposed
                                                                    instead of original names,
                                                                    as values. for example
                                                                    if there is `CarEntity` and `
                                                                    PersonEntity`, it should be
                                                                    like this:
                                                                    `rename=
                                                                    dict(CarEntity=
                                                                         dict(name='new_name'),
                                                                         PersonEntity=
                                                                         dict(age='new_age')`
                                                                    then, the value of `name`
                                                                    column in result will be
                                                                    returned as `new_name` column.
                                                                    and also value of `age` column
                                                                    in result will be returned as
                                                                    'new_age' column. if provided
                                                                    rename columns are not
                                                                    available in result, they
                                                                    will be ignored.

        :note rename: dict[str entity_class_name, dict[str original_column, str new_column]] |
                      dict[str original_column, str new_column]

        :keyword dict[str, list[str]] | list[str] exclude: column names to be excluded from
                                                           result. it could be a list of column
                                                           names. for example:
                                                           `exclude=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity`
                                                           and `PersonEntity`, it should be
                                                           like this:
                                                           `exclude=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                            if provided excluded columns are not
                                                            available in result, they will be
                                                            ignored.

        :note exclude: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.

        :raises InvalidDepthProvidedError: invalid depth provided error.

        :rtype: dict
        """

        all_attributes = None
        relations = None
        requested_columns, rename, excluded_columns = self._extract_conditions(**options)
        readable = options.get('readable', SECURE_TRUE)

        depth = options.get('depth', None)
        if depth is None:
            depth = config_services.get('database', 'conversion', 'default_depth')

        if readable is SECURE_FALSE:
            all_attributes = self.all_attributes
            relations = self.relationships
        else:
            all_attributes = self.all_readable_attributes
            relations = self.exposed_relationships

        requested_relationships = []
        all_attributes = set(all_attributes)
        if len(requested_columns) > 0:
            requested_columns = requested_columns.intersection(all_attributes)
        else:
            requested_columns = all_attributes.difference(excluded_columns)

        result = DTO()
        for col in requested_columns:
            if col in relations:
                requested_relationships.append(col)
            else:
                result[rename.get(col, col)] = getattr(self, col)

        if depth > 0 and len(requested_relationships) > 0:
            if depth > self.MAX_DEPTH:
                raise InvalidDepthProvidedError('Maximum valid "depth" for conversion '
                                                'is [{max_depth}]. provided depth '
                                                '[{invalid_depth}] is invalid.'
                                                .format(max_depth=self.MAX_DEPTH,
                                                        invalid_depth=depth))

            options.update(depth=depth - 1)
            for relation in requested_relationships:
                value = getattr(self, relation)
                new_name = rename.get(relation, relation)
                result[new_name] = None
                if value is not None:
                    if isinstance(value, LIST_TYPES):
                        result[new_name] = []
                        if len(value) > 0:
                            for entity in value:
                                result[new_name].append(entity.to_dict(**options))
                    else:
                        result[new_name] = value.to_dict(**options)

        return result

    def from_dict(self, **kwargs):
        """
        updates the column values of this entity with values in keyword arguments.

        it could fill primary keys, foreign keys, other columns, hybrid properties
        and also relationship properties provided in keyword arguments.
        note that relationship values must be entities. this method could
        not convert relationships which are dict, into entities.

        :keyword SECURE_TRUE | SECURE_FALSE writable: specifies that any column which has
                                                      `allow_write=False` or its name starts
                                                      with underscore `_`, should not be
                                                      populated from given values. this
                                                      is useful if you want to fill an
                                                      entity with keyword arguments passed
                                                      from client and then doing the
                                                      validation. but do not want to expose
                                                      a security risk. especially in update
                                                      operations. defaults to `SECURE_TRUE`
                                                      if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_invalid_column: specifies that if a key is
                                                                   not available in entity
                                                                   columns, do not raise an
                                                                   error. defaults to
                                                                   `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_pk: specifies that any primary key column
                                                       should not be populated with given
                                                       values. this is useful if you want to
                                                       fill an entity with keyword arguments
                                                       passed from client and then doing the
                                                       validation. but do not want to let user
                                                       set primary keys and exposes a security
                                                       risk. especially in update operations.
                                                       defaults to `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_fk: specifies that any foreign key column
                                                       should not be populated with given
                                                       values. this is useful if you want
                                                       to fill an entity with keyword arguments
                                                       passed from client and then doing the
                                                       validation. but do not want to let user
                                                       set foreign keys and exposes a security
                                                       risk. especially in update operations.
                                                       defaults to `SECURE_FALSE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_relationships: specifies that any relationship
                                                                  property should not be populated
                                                                  with given values. defaults to
                                                                  `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE populate_all: specifies that all available values
                                                          must be populated from provided keyword
                                                          arguments. if set to `SECURE_TRUE`, all
                                                          other parameters will be bypassed.
                                                          this is for convenience of usage.
                                                          defaults to `SECURE_FALSE` if not
                                                          provided.

        :raises ColumnNotExistedError: column not existed error.
        """

        ignore_invalid = kwargs.pop('ignore_invalid_column', SECURE_TRUE)
        populate_all = kwargs.pop('populate_all', SECURE_FALSE)
        if populate_all is SECURE_TRUE:
            writable = SECURE_FALSE
            ignore_pk = SECURE_FALSE
            ignore_fk = SECURE_FALSE
            ignore_relationships = SECURE_FALSE
        else:
            writable = kwargs.pop('writable', SECURE_TRUE)
            ignore_pk = kwargs.pop('ignore_pk', SECURE_TRUE)
            ignore_fk = kwargs.pop('ignore_fk', SECURE_FALSE)
            ignore_relationships = kwargs.pop('ignore_relationships', SECURE_TRUE)

        accessible_columns = self.writable_columns
        if writable is SECURE_FALSE:
            accessible_columns = self.all_columns

        accessible_hybrid_properties = self.writable_hybrid_properties
        if writable is SECURE_FALSE:
            accessible_hybrid_properties = self.all_setter_hybrid_properties

        accessible_pk = ()
        if ignore_pk is SECURE_FALSE:
            if writable is SECURE_FALSE:
                accessible_pk = self.primary_key_columns
            else:
                accessible_pk = self.writable_primary_key_columns

        accessible_fk = ()
        if ignore_fk is not SECURE_TRUE:
            if writable is SECURE_FALSE:
                accessible_fk = self.foreign_key_columns
            else:
                accessible_fk = self.writable_foreign_key_columns

        accessible_relationships = ()
        if ignore_relationships is SECURE_FALSE:
            if writable is SECURE_FALSE:
                accessible_relationships = self.relationships
            else:
                accessible_relationships = self.exposed_relationships

        all_accessible_columns = accessible_pk + accessible_fk + \
            accessible_columns + accessible_relationships + accessible_hybrid_properties

        provided_columns = set(kwargs.keys())
        result_columns = set(all_accessible_columns).intersection(provided_columns)
        if ignore_invalid is SECURE_FALSE:
            not_existed = provided_columns.difference(result_columns)
            if len(not_existed) > 0:
                raise ColumnNotExistedError('Provided columns, relationships or properties '
                                            '{columns} are not available in entity [{entity}].'
                                            .format(columns=list(not_existed),
                                                    entity=self.get_fully_qualified_name()))

        for column in result_columns:
            setattr(self, column, kwargs.get(column))

    def set_attribute(self, name, value, silent=True):
        """
        sets the provided value for attribute with given name.

        it only could set value for writable attributes, which are those that
        have `allow_write=True` in their definition (only for columns) and
        their name does not start with underscore `_`.

        this method is implemented to be used in validator package.
        it is not recommended to be used in application code.
        use `from_dict` method instead.

        :param str name: attribute name.
        :param object value: value to be set for attribute.

        :param bool silent: specifies that if there is no attribute
                            with given name, ignore it instead of error.
                            defaults to True.

        :raises ColumnNotExistedError: column not existed error.
        """

        if name in self.all_writable_attributes:
            setattr(self, name, value)
            return

        if silent is False:
            raise ColumnNotExistedError('Provided column, relationship or property '
                                        '[{column}] is not available in entity [{entity}].'
                                        .format(column=name,
                                                entity=self.get_fully_qualified_name()))

    def _extract_conditions(self, **options):
        """
        extracts all conditions available in given options.

        it extracts columns, rename and exclude values.

        :keyword dict[str, list[str]] | list[str] columns: column names to be included in result.
                                                           it could be a list of column names.
                                                           for example:
                                                           `columns=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity` and
                                                           `PersonEntity`, it should be like this:
                                                           `columns=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                           if provided column names are not
                                                           available in result, they will
                                                           be ignored.

        :note columns: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :keyword dict[str, dict[str, str]] | dict[str, str] rename: column names that must be
                                                                    renamed in the result.
                                                                    it could be a dict with keys
                                                                    as original column names and
                                                                    values as new column names
                                                                    that should be exposed instead
                                                                    of original column names.
                                                                    for example:
                                                                    `rename=dict(age='new_age',
                                                                                 name='new_name')`
                                                                    but if you want to include
                                                                    relationships, then you must
                                                                    provide a dict containing
                                                                    entity class name as key and
                                                                    for value, another dict
                                                                    containing original column
                                                                    names as keys, and column
                                                                    names that must be exposed
                                                                    instead of original names,
                                                                    as values. for example
                                                                    if there is `CarEntity` and
                                                                    `PersonEntity`, it should be
                                                                    like this:
                                                                    `rename=
                                                                    dict(CarEntity=
                                                                         dict(name='new_name'),
                                                                         PersonEntity=
                                                                         dict(age='new_age')`
                                                                    then, the value of `name`
                                                                    column in result will be
                                                                    returned as `new_name` column.
                                                                    and also value of `age` column
                                                                    in result will be returned as
                                                                    'new_age' column. if provided
                                                                    rename columns are not
                                                                    available in result, they
                                                                    will be ignored.

        :note rename: dict[str entity_class_name, dict[str original_column, str new_column]] |
                      dict[str original_column, str new_column]

        :keyword dict[str, list[str]] | list[str] exclude: column names to be excluded from
                                                           result. it could be a list of column
                                                           names. for example:
                                                           `exclude=['id', 'name', 'age']`
                                                           but if you want to include
                                                           relationships, then columns for each
                                                           entity must be provided in a key for
                                                           that entity class name.
                                                           for example if there is `CarEntity`
                                                           and `PersonEntity`, it should be
                                                           like this:
                                                           `exclude=dict(CarEntity=
                                                                         ['id', 'name'],
                                                                         PersonEntity=
                                                                         ['id', 'age'])`
                                                            if provided excluded columns are not
                                                            available in result, they will be
                                                            ignored.

        :note exclude: dict[str entity_class_name, list[str column_name]] | list[str column_name]

        :returns: tuple[set[str column_name],
                        dict[str original_column, str new_column],
                        set[str excluded_column]]

        :rtype: tuple[set[str], dict[str, str], set[str]]
        """

        columns = options.get('columns', None)
        rename = options.get('rename', None)
        exclude = options.get('exclude', None)

        if isinstance(columns, dict):
            columns = columns.get(self.get_class_name(), None)

        if isinstance(rename, dict) and len(rename) > 0 and \
                isinstance(list(rename.values())[0], dict):

            rename = rename.get(self.get_class_name(), None)

        if isinstance(exclude, dict):
            exclude = exclude.get(self.get_class_name(), None)

        if columns is None:
            columns = []

        if rename is None:
            rename = {}

        if exclude is None:
            exclude = []

        columns = set(columns)
        exclude = set(exclude)
        return columns.difference(exclude), rename, exclude


class MagicMethodMixin(ModelMixinBase):
    """
    magic method mixin class.

    this class adds different magic method implementations to its subclasses.
    """

    def __eq__(self, other):
        """
        gets the equality comparison result.

        first, it compares primary keys if they have value in both entities
        and both entities have a common root parent.
        otherwise it compares them using python default memory location comparison.

        :param CoreEntity other: other entity to compare for equality.

        :rtype: bool
        """

        if isinstance(other, self.root_base_class):
            if self._is_primary_key_comparable(self.primary_key()) is True:
                return self.primary_key() == other.primary_key()
            else:
                return self is other

        return False

    def __ne__(self, other):
        """
        gets the not equality comparison result.

        :param CoreEntity other: other entity to compare for not equality.

        :rtype: bool
        """

        return not self == other

    def __hash__(self):
        """
        gets the hash of current entity.

        if the entity has valid primary key values, it will be considered
        in hash generation. otherwise it falls back to general hash generation
        based on python defaults.

        :rtype: int
        """

        if self._is_primary_key_comparable(self.primary_key()) is True:
            return hash('{root_base}.{pk}'.format(root_base=self.root_base_class,
                                                  pk=self.primary_key()))

        return super().__hash__()

    def __str__(self):
        """
        gets the string representation of current entity.

        :rtype: str
        """

        return '{fullname} -> {pk}'.format(fullname=self.get_fully_qualified_name(),
                                           pk=self.primary_key())

    @class_property
    @fast_cache
    def root_base_class(cls):
        """
        gets root base class of this entity.

        root base class is the class which is direct subclass
        of declarative base class (which by default is CoreEntity)
        in inheritance hierarchy.

        for example if you use pyrin's default CoreEntity as your base model:
        {CoreEntity -> BaseEntity, A -> CoreEntity, B -> A, C -> B}
        then, root base class of A, B and C is A.

        if you implement a new base class named MyNewDeclarativeBase as base model:
        {MyNewDeclarativeBase -> BaseEntity, A -> MyNewDeclarativeBase, B -> A, C -> B}
        then, root base class of A, B and C is A.

        the inheritance rule also supports multi-branch hierarchy. for example:
        {CoreEntity -> BaseEntity, A -> CoreEntity, B -> A, C -> A}
        then, root base class of A, B and C is A.

        root base class will be calculated once and cached per entity type.

        :rtype: type
        """

        bases = inspect.getmro(cls)
        root_base_entity_index = bases.index(cls.declarative_base_class) - 1
        base = bases[root_base_entity_index]
        return base

    @class_property
    def declarative_base_class(cls):
        """
        gets declarative base class of application.

        :rtype: type
        """

        base = cls._get_declarative_base_class()
        if base is None:
            bases = inspect.getmro(cls)
            base_entity_index = bases.index(cls._base_entity_class)
            potential_declarative_bases = bases[0:base_entity_index]
            base = cls._extract_declarative_base(potential_declarative_bases)
            cls._set_declarative_base_class(base)

        return base

    @classmethod
    def _extract_declarative_base(cls, types):
        """
        extracts the first declarative base found from given types.

        returns None if not found.

        :param tuple[type] types: class types to extract declarative base from them.

        :rtype: type
        """

        for item in types:
            try:
                sqla_inspect(item)
            except NoInspectionAvailable:
                return item

        return None

    @classmethod
    def _set_declarative_base_class(cls, declarative_base_class):
        """
        sets declarative base class of application.

        the value will be set and shared for all entities because there
        should be only one declarative base class.

        :param type declarative_base_class: a class type of application declarative base class.
                                            by default, it would be `CoreEntity` class.
        """

        cls._validate_declarative_base_class(declarative_base_class)
        MagicMethodMixin._declarative_base_class = declarative_base_class

    @classmethod
    def _get_declarative_base_class(cls):
        """
        gets declarative base class of application.

        returns None if it's not set.

        :rtype: type
        """

        return getattr(MagicMethodMixin, '_declarative_base_class', None)

    @class_property
    @abstractmethod
    def _base_entity_class(cls):
        """
        gets base entity class of application.

        this method must be overridden in `BaseEntity` class.
        it should return type of `BaseEntity` class itself.
        this method is required to overcome circular dependency problem as mixin
        module could not import `BaseEntity` because `BaseEntity` itself references
        to mixin module. and also we could not inject `BaseEntity` dependency through
        `__init__()` method of `MagicMethodMixin` class, because sqlalchemy does not
        call `__init__()` method of entities for populating database results, so
        `__init__()` call is not guaranteed and will only take place on user code.
        so we have to define this method to get `BaseEntity` type here.
        and this is more beautiful then importing `BaseEntity` inside a method
        of `MagicMethodMixin` class.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type
        """

        raise CoreNotImplementedError()

    @classmethod
    def _validate_declarative_base_class(cls, declarative_base_class):
        """
        validates the given declarative base class.

        :param type declarative_base_class: a class type of application declarative base class.
                                            by default, it would be `CoreEntity` class.

        :raises InvalidDeclarativeBaseTypeError: invalid declarative base type error.
        """

        if declarative_base_class is None or not inspect.isclass(declarative_base_class):
            raise InvalidDeclarativeBaseTypeError('Input parameter [{declarative}] '
                                                  'is not a class.'
                                                  .format(declarative=declarative_base_class))

        if not issubclass(declarative_base_class, cls._base_entity_class):
            raise InvalidDeclarativeBaseTypeError('Input parameter [{declarative}] '
                                                  'in not a subclass of [{base}].'
                                                  .format(declarative=declarative_base_class,
                                                          base=cls._base_entity_class))

        if declarative_base_class is not model_services.get_declarative_base():
            print_warning('You have implemented a new declarative base type [{new}] '
                          'in your application. to make everything works as expected '
                          'you must override "pyrin.database.model.ModelManager.'
                          'get_declarative_base()" method in your application inside '
                          '"database.model" package. for more information on how to do '
                          'that or how to ignore it, see the documentation of specified '
                          'method.'.format(new=declarative_base_class))

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.root_base_class
        super().populate_cache()


class QueryMixin:
    """
    query mixin class.

    this class adds query method to its subclasses.
    the query method will always use the correct session
    based on request context availability.
    """

    @classmethod
    def query(cls, *entities, **options):
        """
        gets the query object to perform queries on it.

        it always uses the correct session based on request context availability.

        :param BaseEntity entities: entities or columns to use them in query.
                                    if not provided, it uses all columns of this entity.

        :keyword type | tuple[type] scope: class type of the entities that this
                                           query instance will work on. if the
                                           query is working on multiple entities,
                                           this value must be a tuple of all class
                                           types of that entities.

                                           for example: if you set
                                           `entities=SomeEntity.id, AnotherEntity.name`
                                           you should leave `scope=None` to skip validation
                                           or you could set
                                           `scope=(SomeEntity, AnotherEntity)`
                                           this way validation succeeds, but if
                                           you set `scope=SomeEntity`
                                           then the query will not be executed
                                           and an error will be raised.

        :raises ColumnsOutOfScopeError: columns out of scope error.

        :rtype: CoreQuery
        """

        store = get_current_store()
        used_entities = entities
        if entities is None or len(entities) <= 0:
            used_entities = (cls,)

        return store.query(*used_entities, **options)


class CRUDMixin:
    """
    crud mixin class.

    this class adds CRUD functionalities to is subclasses.
    it includes save, update and delete.
    it uses the correct session based on request context availability.
    """

    def save(self, flush=False):
        """
        saves the current entity.

        :param bool flush: flush changes to database at the end.
                           defaults to False if not provided.
                           this would be helpful if you need to get
                           the autogenerated values by database such as sequences.

        :rtype: BaseEntity
        """

        store = get_current_store()
        store.add(self)

        if flush is True:
            store.flush()

        return self

    def update(self, flush=False, **kwargs):
        """
        updates the column values of this entity with values in keyword arguments.

        then persists changes into database.
        it could fill primary keys, foreign keys, other columns, hybrid properties
        and also relationship properties provided in keyword arguments.
        note that relationship values must be entities. this method could
        not convert relationships which are dict, into entities.

        :param bool flush: flush changes to database at the end.
                           defaults to False if not provided.
                           this would be helpful if you need to get
                           the autogenerated values by database such as sequences.

        :keyword SECURE_TRUE | SECURE_FALSE writable: specifies that any column which has
                                                      `allow_write=False` or its name starts
                                                      with underscore `_`, should not be
                                                      populated from given values. this
                                                      is useful if you want to fill an
                                                      entity with keyword arguments passed
                                                      from client and then doing the
                                                      validation. but do not want to expose
                                                      a security risk. especially in update
                                                      operations. defaults to `SECURE_TRUE`
                                                      if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_invalid_column: specifies that if a key is
                                                                   not available in entity
                                                                   columns, do not raise an
                                                                   error. defaults to
                                                                   `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_pk: specifies that any primary key column
                                                       should not be populated with given
                                                       values. this is useful if you want to
                                                       fill an entity with keyword arguments
                                                       passed from client and then doing the
                                                       validation. but do not want to let user
                                                       set primary keys and exposes a security
                                                       risk. especially in update operations.
                                                       defaults to `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_fk: specifies that any foreign key column
                                                       should not be populated with given
                                                       values. this is useful if you want
                                                       to fill an entity with keyword arguments
                                                       passed from client and then doing the
                                                       validation. but do not want to let user
                                                       set foreign keys and exposes a security
                                                       risk. especially in update operations.
                                                       defaults to `SECURE_FALSE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE ignore_relationships: specifies that any relationship
                                                                  property should not be populated
                                                                  with given values. defaults to
                                                                  `SECURE_TRUE` if not provided.

        :keyword SECURE_TRUE | SECURE_FALSE populate_all: specifies that all available values
                                                          must be populated from provided keyword
                                                          arguments. if set to `SECURE_TRUE`, all
                                                          other parameters will be bypassed.
                                                          this is for convenience of usage.
                                                          defaults to `SECURE_FALSE` if not
                                                          provided.

        :keyword SECURE_TRUE | SECURE_FALSE prefetch_complex_defaults: specifies that all columns
                                                                       that have default values
                                                                       for update must fetch their
                                                                       default value if no value
                                                                       is provided for them in
                                                                       `kwargs`.
                                                                       note that scalar default
                                                                       values will always be
                                                                       fetched and this option is
                                                                       only for sequence or
                                                                       callable default values.
                                                                       defaults to `SECURE_TRUE`
                                                                       if not provided.

        :note prefetch_complex_defaults: for callable defaults we have to pass
                                         `ExecutionContext` as None because there is no
                                         such context in prefetch time. so if your callable
                                         default actually needs a valid context, you can not
                                         prefetch its value and must disable prefetching by
                                         passing `prefetch_complex_defaults=SECURE_FALSE`.
                                         otherwise unexpected behavior may occur.

        :raises ColumnNotExistedError: column not existed error.

        :rtype: BaseEntity
        """

        self.from_dict(**kwargs)
        self.prefetch_update_defaults(**kwargs)
        return self.save(flush=flush)

    def delete(self, flush=False):
        """
        deletes the current entity.

        :param bool flush: flush changes to database at the end.
                           defaults to False if not provided.
                           this would be helpful if you need to get
                           the autogenerated values by database such as sequences.
        """

        store = get_current_store()
        store.delete(self)

        if flush is True:
            store.flush()


class MetadataMixin:
    """
    metadata mixin class.

    this class provides a simple and extensible api for declarative models configuration.
    """

    # table name
    _table = None

    # table args
    _schema = None
    _extend_existing = None

    # specifies that this entity is abstract.
    # abstract entities does not represent any table on database.
    _abstract = False

    # a list of entity column instances or table column names to
    # create unique constraint for them.
    # for example:
    # _unique_on = name -> single column.
    # _unique_on = name, age -> multiple columns in one constraint.
    # _unique_on = identity, (name, age), username -> multiple columns in three constraints.
    _unique_on = None

    # mapper args
    _polymorphic_on = None
    _polymorphic_identity = None
    _concrete = None

    @classmethod
    @fast_cache
    def _get_unique_constraints(cls):
        """
        gets required unique constraints for this entity.

        :rtype: tuple[UniqueConstraint]
        """

        if cls._unique_on is None:
            return tuple()

        uniques = misc_utils.make_iterable(cls._unique_on, tuple)
        is_multiple = False
        for item in uniques:
            if isinstance(item, LIST_TYPES):
                is_multiple = True
                break

        constraints = []
        if is_multiple is False:
            constraints.append(UniqueConstraint(*uniques))
        else:
            for item in uniques:
                iterable_item = misc_utils.make_iterable(item, tuple)
                constraints.append(UniqueConstraint(*iterable_item))

        return tuple(constraints)

    @declared_attr
    def __abstract__(cls):
        """
        gets a value indicating that this entity is abstract.

        abstract entities does not represent any table on database.

        :rtype: bool
        """

        return cls._abstract

    @declared_attr
    def __tablename__(cls):
        """
        gets the table name of current entity type.

        it returns the value of `_table` class attribute of entity.

        :rtype: str
        """

        return cls._table

    @declared_attr
    @fast_cache
    def __table_args__(cls):
        """
        gets the table args of current entity type.

        it returns a dict or tuple of all configured table args.

        for example: {'schema': 'database_name.schema_name',
                      'extend_existing': True}

        :rtype: dict | tuple
        """

        table_args = dict()
        if cls._schema is not None:
            table_args.update(schema=cls._schema)

        if cls._extend_existing is not None:
            table_args.update(extend_existing=cls._extend_existing)

        unique_constraints = cls._get_unique_constraints()
        extra_args = cls._customize_table_args(table_args)
        if extra_args is not None:
            extra_args = misc_utils.make_iterable(extra_args, tuple)
        else:
            extra_args = tuple()

        all_args = unique_constraints + extra_args
        if len(all_args) <= 0:
            return table_args

        return all_args + (table_args,)

    @declared_attr
    @fast_cache
    def __mapper_args__(cls):
        """
        gets the mapper args of current entity type.

        :rtype: dict
        """

        mapper_args = dict()
        if cls._polymorphic_on is not None:
            mapper_args.update(polymorphic_on=cls._polymorphic_on)

        if cls._polymorphic_identity is not None:
            mapper_args.update(polymorphic_identity=cls._polymorphic_identity)

        if cls._concrete is not None:
            mapper_args.update(concrete=cls._concrete)

        cls._customize_mapper_args(mapper_args)
        return mapper_args

    @classmethod
    def _customize_table_args(cls, table_args):
        """
        customizes different table args for current entity type.

        this method is intended to be overridden by subclasses to customize
        table args per entity type if the required customization needs extra work.
        it must modify input dict values in-place if required.
        if other table args must be added (ex. UniqueConstraint or CheckConstraint ...)
        it must return those as a tuple. it could also return a single object as
        extra table arg (ex. a single UniqueConstraint).
        if no changes are required this method must return None.

        :param dict table_args: a dict containing different table args.
                                any changes to this dict must be done in-place.

        :rtype: tuple | object
        """

        return None

    @classmethod
    def _customize_mapper_args(cls, mapper_args):
        """
        customizes different mapper args for current entity type.

        this method is intended to be overridden by subclasses to customize
        mapper args per entity type if the required customization needs extra work.
        it must modify values of input dict in-place if required.

        :param dict mapper_args: a dict containing different mapper args.
        """
        pass

    @class_property
    def table_name(cls):
        """
        gets the table name that this entity represents in database.

        :rtype: str
        """

        return cls._table

    @class_property
    def table_schema(cls):
        """
        gets the table schema that this entity represents in database.

        it might be `None` if schema has not been set for this entity.

        :rtype: str
        """

        return cls._schema

    @class_property
    def table_fullname(cls):
        """
        gets the table fullname that this entity represents in database.

        fullname is `schema.table_name` if schema is available, otherwise it
        defaults to `table_name`.

        :rtype: str
        """

        if cls.table_schema is not None:
            return '{schema}.{name}'.format(schema=cls.table_schema, name=cls.table_name)
        else:
            return cls.table_name

    @class_property
    def extend_existing(cls):
        """
        gets a value indicating that this entity extends another entity.

        :rtype: bool
        """

        return cls._extend_existing


class ModelCacheMixin(TypedCacheMixin):
    """
    model cache mixin class.

    this class adds caching support to its subclasses.
    """

    _container = {}


class DefaultPrefetchMixin(ModelMixinBase):
    """
    default prefetch mixin class.

    this class adds support to prefetch columns with default or onupdate
    values without flush or commit.
    """

    @classmethod
    def _get_column_attribute(cls, name):
        """
        gets column attribute with given name.

        :param str name: column name.

        :rtype: CoreColumn
        """

        return cls.all_column_attributes.get(name)

    def _get_insert_default_value(self, column):
        """
        gets the insert default value for given column.

        :param str column: column name.

        :returns: object
        """

        attribute = self._get_column_attribute(column)
        return self._fetch_default_value(attribute.default)

    def _get_update_default_value(self, column):
        """
        gets the update default value for given column.

        :param str column: column name.

        :returns: object
        """

        attribute = self._get_column_attribute(column)
        return self._fetch_default_value(attribute.onupdate)

    def _set_insert_default(self, column):
        """
        sets the insert default value to the given column.

        :param str column: column name.
        """

        value = self._get_insert_default_value(column)
        setattr(self, column, value)

    def _set_update_default(self, column):
        """
        sets the update default value to the given column.

        :param str column: column name.
        """

        value = self._get_update_default_value(column)
        setattr(self, column, value)

    @classmethod
    def _fetch_default_value(cls, default):
        """
        fetches the value from given default clause.

        note that for callable defaults we have to pass `ExecutionContext` as None
        because there is no such context in prefetch time. so if your callable default
        actually needs a valid context, you can not prefetch its value and must disable
        prefetching by passing `prefetch_complex_defaults=False` in corresponding method.
        otherwise unexpected behavior may occur.

        :param ColumnDefault default: column default instance.

        :returns: object
        """

        # the order of if conditions must be exactly this way. because
        # 'Sequence' object does not have the other attributes.
        if default.is_sequence:
            store = get_current_store()
            return store.execute(default.next_value()).scalar()
        elif default.is_scalar:
            return default.arg
        elif default.is_callable:
            return default.arg(None)

        return cls._fetch_default_value_extended(default)

    @classmethod
    @abstractmethod
    def _fetch_default_value_extended(cls, default):
        """
        fetches the value from given default clause.

        this method is intended to be overridden in subclasses to modify or extend
        prefetching default values. for example to support `ClauseElement` defaults.

        :param ColumnDefault default: column default instance.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @class_property
    @fast_cache
    def columns_with_scalar_insert_default(cls):
        """
        gets column names that have scalar default values for insert.

        :rtype: tuple[str]
        """

        all_columns = cls.primary_key_columns + cls.all_columns + cls.foreign_key_columns
        result = []
        for name in all_columns:
            attribute = cls._get_column_attribute(name)
            if attribute.default is not None and \
                    not attribute.default.is_sequence and attribute.default.is_scalar:
                result.append(name)

        return tuple(result)

    @class_property
    @fast_cache
    def columns_with_complex_insert_default(cls):
        """
        gets column names that have callable or sequence default values for insert.

        :rtype: tuple[str]
        """

        all_columns = cls.primary_key_columns + cls.all_columns + cls.foreign_key_columns
        result = []
        for name in all_columns:
            attribute = cls._get_column_attribute(name)
            if attribute.default is not None and (attribute.default.is_sequence or
                                                  attribute.default.is_callable):
                result.append(name)

        return tuple(result)

    @class_property
    @fast_cache
    def columns_with_scalar_update_default(cls):
        """
        gets column names that have scalar default values for update.

        :rtype: tuple[str]
        """

        all_columns = cls.all_columns + cls.foreign_key_columns
        result = []
        for name in all_columns:
            attribute = cls._get_column_attribute(name)
            if attribute.onupdate is not None and \
                    not attribute.onupdate.is_sequence and attribute.onupdate.is_scalar:
                result.append(name)

        return tuple(result)

    @class_property
    @fast_cache
    def columns_with_complex_update_default(cls):
        """
        gets column names that have callable or sequence default values for update.

        :rtype: tuple[str]
        """

        all_columns = cls.all_columns + cls.foreign_key_columns
        result = []
        for name in all_columns:
            attribute = cls._get_column_attribute(name)
            if attribute.onupdate is not None and (attribute.onupdate.is_sequence or
                                                   attribute.onupdate.is_callable):
                result.append(name)

        return tuple(result)

    def prefetch_insert_defaults(self, **kwargs):
        """
        prefetches all columns that have default values for insert.

        and their name is not in provided inputs or their value is None.

        :keyword SECURE_TRUE | SECURE_FALSE prefetch_complex_defaults: specifies that all columns
                                                                       that have default values
                                                                       for insert must fetch their
                                                                       default value if no value
                                                                       is provided for them in
                                                                       `kwargs`.
                                                                       note that scalar default
                                                                       values will always be
                                                                       fetched and this option is
                                                                       only for sequence or
                                                                       callable default values.
                                                                       defaults to `SECURE_TRUE`
                                                                       if not provided.

        :note prefetch_complex_defaults: for callable defaults we have to pass
                                         `ExecutionContext` as None because there is no
                                         such context in prefetch time. so if your callable
                                         default actually needs a valid context, you can not
                                         prefetch its value and must disable prefetching by
                                         passing `prefetch_complex_defaults=SECURE_FALSE`.
                                         otherwise unexpected behavior may occur.
        """

        fetch_complex_defaults = kwargs.pop('prefetch_complex_defaults', SECURE_TRUE)
        self._prefetch_insert_defaults(self.columns_with_scalar_insert_default, **kwargs)

        if fetch_complex_defaults is SECURE_TRUE:
            self._prefetch_insert_defaults(self.columns_with_complex_insert_default, **kwargs)

    def _prefetch_insert_defaults(self, columns, **kwargs):
        """
        prefetches all columns that have default values for insert.

        and their name is not in provided inputs or their value is None.

        :param tuple[str] columns: column names that have insert default values.
        """

        for item in columns:
            provided = kwargs.get(item)
            if provided is None:
                self._set_insert_default(item)

    def prefetch_update_defaults(self, **kwargs):
        """
        prefetches all columns that have default values for update.

        and their name is not in provided inputs or their value is None.

        :keyword SECURE_TRUE | SECURE_FALSE prefetch_complex_defaults: specifies that all columns
                                                                       that have default values
                                                                       for update must fetch their
                                                                       default value if no value
                                                                       is provided for them in
                                                                       `kwargs`.
                                                                       note that scalar default
                                                                       values will always be
                                                                       fetched and this option is
                                                                       only for sequence or
                                                                       callable default values.
                                                                       defaults to `SECURE_TRUE`
                                                                       if not provided.

        :note prefetch_complex_defaults: for callable defaults we have to pass
                                         `ExecutionContext` as None because there is no
                                         such context in prefetch time. so if your callable
                                         default actually needs a valid context, you can not
                                         prefetch its value and must disable prefetching by
                                         passing `prefetch_complex_defaults=SECURE_FALSE`.
                                         otherwise unexpected behavior may occur.
        """

        fetch_complex_defaults = kwargs.pop('prefetch_complex_defaults', SECURE_TRUE)
        self._prefetch_update_defaults(self.columns_with_scalar_update_default, **kwargs)

        if fetch_complex_defaults is SECURE_TRUE:
            self._prefetch_update_defaults(self.columns_with_complex_update_default, **kwargs)

    def _prefetch_update_defaults(self, columns, **kwargs):
        """
        prefetches all columns that have default values for update.

        and their name is not in provided inputs or their value is None.

        :param tuple[str] columns: column names that have update default values.
        """

        for item in columns:
            provided = kwargs.get(item)
            if provided is None:
                self._set_update_default(item)

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.columns_with_scalar_insert_default
        temp = cls.columns_with_complex_insert_default
        temp = cls.columns_with_scalar_update_default
        temp = cls.columns_with_complex_update_default
        super().populate_cache()


class OrderingMixin(ModelMixinBase):
    """
    ordering mixin class.

    this class adds functionalities about ordering to its subclasses.
    """

    # a list of entity column instances which are allowed to be used in order by criterion.
    # defaults to all columns of entity if not provided. if set to an empty list ordering
    # will be disabled for all columns.
    # for example:
    # _ordering_columns = None -> allow ordering on all columns (default).
    # _ordering_columns = [] -> disable ordering on all columns.
    # _ordering_columns = name -> allow ordering only on name column.
    # _ordering_columns = age, name, family -> allow ordering on three columns.
    _ordering_columns = None

    @class_property
    @fast_cache
    def ordering_column_names(cls):
        """
        gets all column attribute names which are allowed in order by.

        the result will be calculated once and cached per entity type.

        :raises InvalidOrderingColumnTypeError: invalid ordering column type error.
        :raises ColumnNotExistedError: column not existed error.

        :rtype: tuple[str]
        """

        if cls._ordering_columns is None:
            return cls.all_columns + cls.primary_key_columns + cls.foreign_key_columns

        iterable_columns = misc_utils.make_iterable(cls._ordering_columns, tuple)
        if len(iterable_columns) <= 0:
            return tuple()

        result = []
        for item in iterable_columns:
            if not isinstance(item, CoreColumn):
                raise InvalidOrderingColumnTypeError('Ordering columns on entity [{entity}] '
                                                     'must be an instance of [{column}].'
                                                     .format(entity=cls, column=CoreColumn))

            name = cls.all_reverse_column_attributes.get(item)
            if name is None:
                raise ColumnNotExistedError('Ordering column [{column}] does '
                                            'not exist in entity [{entity}].'
                                            .format(column=item, entity=cls))

            result.append(name)

        return tuple(result)

    @classmethod
    def get_ordering_criterion(cls, *columns, ignore_invalid=True):
        """
        gets required ordering criterion for given column names.

        default ordering is ascending, but it could be changed to descending
        by prefixing `-` to column names.

        for example:

        name, +age -> ordering for name and age columns both ascending.
        name, -age -> ordering for name ascending and age descending.

        :param str columns: column names to get their ordering criterion.
                            they must be the column attribute names of
                            this entity.

        :param bool ignore_invalid: specifies that if provided columns are
                                    not valid, ignore them instead of raising
                                    an error. defaults to True.

        :raises InvalidOrderingColumnError: invalid ordering column error.

        :rtype: tuple[UnaryExpression]
        """

        result = []
        for item in columns:
            if sqlalchemy_utils.is_valid_column_name(item):
                name, order_type = sqlalchemy_utils.get_ordering_info(item)
                attribute = cls._get_column_attribute(name)
                is_allowed = name in cls.ordering_column_names
                if (attribute is None or not is_allowed) and ignore_invalid is False:
                    raise InvalidOrderingColumnError(_('Column [{name}] is not valid '
                                                       'for ordering.').format(name=name))
                elif attribute is not None and is_allowed:
                    result.append(order_type(attribute))

        return tuple(result)

    @classmethod
    def populate_cache(cls):
        """
        populates all related caches.
        """

        temp = cls.ordering_column_names
        super().populate_cache()


class CreateHistoryMixin:
    """
    create history mixin class.

    this class adds `created_on` column into its subclasses.
    """

    created_on = TimeStampColumn(name='created_on', nullable=False, validated=True,
                                 validated_find=False, validated_range=True,
                                 default=datetime_services.now, allow_write=False)


class UpdateHistoryMixin:
    """
    update history mixin class.

    this class adds `modified_on` column into its subclasses.
    """

    modified_on = TimeStampColumn(name='modified_on', nullable=True, validated=True,
                                  validated_find=False, validated_range=True,
                                  onupdate=datetime_services.now, allow_write=False)


class HistoryMixin(CreateHistoryMixin, UpdateHistoryMixin):
    """
    history mixin class.

    this class adds `created_on` and `modified_on` columns into its subclasses.
    """
    pass



# -*- coding: utf-8 -*-
"""
cli arguments module.
"""

from abc import ABC, abstractmethod

from flask import json as flask_json

from pyrin.core.globals import LIST_TYPES
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import CoreObject, DTO
from pyrin.cli.exceptions import ArgumentParamNameIsRequiredError, \
    PositionalArgumentIndexError, BooleanArgumentValueError, \
    KeywordArgumentCLIOptionNameError, MappingArgumentParamValueToCLIMapRequiredError, \
    KeywordArgumentCLIOptionNameRequiredError


class ArgumentBase(CoreObject, ABC):
    """
    argument base class.
    """

    def __init__(self, param_name, default=None):
        """
        initializes an instance of ArgumentBase.

        :param str param_name: param name presented in method signature.

        :param list[object] | object default: default value to be emitted to
                                              cli if this param is not available.
                                              if set to None, this param will not
                                              be emitted at all.
                                              defaults to None if not provided.

        :raises ArgumentParamNameIsRequiredError: argument param name is required error.
        """

        super().__init__()

        if param_name in (None, '') or param_name.isspace():
            raise ArgumentParamNameIsRequiredError('Argument "param_name" '
                                                   'must be provided.')
        self._param_name = param_name
        self._default = default

    def __str__(self):
        """
        gets the string representation of current argument.

        :rtype: str
        """

        return '{base} [{param}]'.format(base=super().__str__(), param=self.param_name)

    def _convert_result(self, value):
        """
        converts the given value to required type.

        this method is intended to be overridden in subclasses if they
        need the result to be of a specific type.

        :param list[object] | object value: value to be converted.

        :rtype: list[object] | object
        """

        return value

    def get_representation(self, value, is_default=False):
        """
        gets the representation of current cli option based on given input.

        it could be None if the value should not be emitted to cli.

        :param list[object] | object value: value of method input
                                            that should be represented
                                            in cli.

        :param bool is_default: specifies that input value is the default
                                value of this argument. it might be needed
                                for some subclasses to differentiate between
                                default value and other values.
                                defaults to False if not provided.

        :rtype: list[object] | object
        """

        if isinstance(value, (set, tuple)):
            value = list(value)

        result = self._get_representation(value, is_default)

        if result is None and is_default is False:
            return self.get_representation(self._default, True)

        return self._convert_result(result)

    @property
    def default(self):
        """
        gets the default value of this argument.

        :rtype: list[object] | object
        """

        return self._default

    @property
    def param_name(self):
        """
        gets the param name of this argument.

        :rtype: str
        """

        return self._param_name

    @abstractmethod
    def _get_representation(self, value, is_default=False):
        """
        gets the representation of current cli option based on given input.

        it could be None if the value should not be emitted to cli.

        :param list[object] | object value: value of method input that
                                            should be represented in cli.

        :param bool is_default: specifies that input value is the default
                                value of this argument. it might be needed
                                for some subclasses to differentiate between
                                default value and other values.
                                defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: list[object] | object
        """

        raise CoreNotImplementedError()


class KeywordArgumentBase(ArgumentBase):
    """
    keyword argument base class.
    """

    def __init__(self, param_name, cli_option_name=None, default=None):
        """
        initializes an instance of KeywordArgumentBase.

        :param str param_name: param name presented in method signature.

        :param str cli_option_name: relevant cli option name to `param_name`.
                                    for example `--message` flag of alembic
                                    must be present with the message value
                                    itself. defaults to None if not provided.

        :param list[object] | object default: default value to be emitted to
                                              cli if this param is not available.
                                              if set to None, this param will not
                                              be emitted at all.
                                              defaults to None if not provided.

        :raises ArgumentParamNameIsRequiredError: argument param name is required error.
        :raises KeywordArgumentCLIOptionNameError: keyword argument cli option name error.
        """

        super().__init__(param_name, default)

        if cli_option_name is not None and \
                (cli_option_name.isspace() or cli_option_name == ''):
            raise KeywordArgumentCLIOptionNameError('Keyword argument "cli_option_name" '
                                                    'could be None or have a valid '
                                                    'value. blank is not valid.')

        self._cli_option_name = cli_option_name

    def _get_representation(self, value, is_default=False):
        """
        gets the representation of current cli option based on given input.

        it could be None if the value should not be emitted to cli.

        :param list[object] | object value: value of method input that
                                            should be represented in cli.

        :param bool is_default: specifies that input value is the default
                                value of this argument. it might be needed
                                for some subclasses to differentiate between
                                default value and other values.
                                defaults to False if not provided.

        :rtype: list[object] | object
        """

        return self._merge_inputs(value)

    def _merge_inputs(self, value):
        """
        merges the cli option name and given value to be emitted to cli.

        note that the values of list type, must be emitted separately
        using the `cli_option_name`. for example if value = [1, 2, 3]
        and `cli_option_name` is `--num`, then it should be emitted to
        cli like this: ['--num', 1, '--num', 2, '--num', 3]
        if the cli handler supports multiple values with single name
        (for example pybabel commands), you should use
        `CompositeKeywordArgument` class instead.

        :param list[object] | object value: value to be emitted to cli.

        :rtype: list[object] | object
        """

        if value is None:
            return None

        if self._cli_option_name is not None:
            result = []
            if isinstance(value, LIST_TYPES):
                for item in value:
                    result.extend([self._cli_option_name, item])
            else:
                result.extend([self._cli_option_name, value])
            return result

        return value


class PositionalArgument(ArgumentBase):
    """
    positional argument class.

    this class must be used for positional cli arguments.
    positional arguments are those that will be emitted after
    command name at specified index and without any argument
    name, for example: `command value1`
    """

    def __init__(self, param_name, index, default=None, **options):
        """
        initializes an instance of PositionalArgument.

        :param str param_name: param name presented in method signature.
        :param int index: zero based index of this param in cli command inputs.

        :param list[object] | object default: default value to be emitted to
                                              cli if this param is not available.
                                              if set to None, this param will not
                                              be emitted at all.
                                              defaults to None if not provided.

        :keyword bool validate_index: specifies that index of this argument
                                      must be validated. it could be helpful
                                      to set this to False when there are multiple
                                      arguments with the same index that will appear
                                      in different situations.
                                      defaults to True if not provided.

        :raises ArgumentParamNameIsRequiredError: argument param name is required error.

        :raises PositionalArgumentIndexError: positional argument index error.
        """

        super().__init__(param_name, default)

        if index is None or index < 0:
            raise PositionalArgumentIndexError('Positional argument "index" '
                                               'could not be less than zero.')
        self._index = index
        self._validate_index = options.get('validate_index', True)

    def _get_representation(self, value, is_default=False):
        """
        gets the representation of current cli option based on given input.

        it could be None if the value should not be emitted to cli.

        :param list[object] | object value: value of method input that
                                            should be represented in cli.

        :param bool is_default: specifies that input value is the default
                                value of this argument. it might be needed
                                for some subclasses to differentiate between
                                default value and other values.
                                defaults to False if not provided.

        :rtype: list[object] | object
        """

        return value

    @property
    def index(self):
        """
        gets the index in which this param must be emitted to cli command.

        :rtype: int
        """

        return self._index

    @property
    def validate_index(self):
        """
        gets a value indicating that index of this argument must be validated.

        :rtype: bool
        """

        return self._validate_index


class CompositePositionalArgument(PositionalArgument):
    """
    composite positional argument class.

    this class must be used for composite positional cli arguments.
    composite positional arguments are those that will be emitted after
    command name at specified index and without any argument name, and the
    value could be a list.
    for example: `command value1` or `command value1 value2 value3`
    """

    def __init__(self, param_name, index, default=None,
                 separator=None, **options):
        """
        initializes an instance of CompositePositionalArgument.

        :param str param_name: param name presented in method signature.
        :param int index: zero based index of this param in cli command inputs.

        :param list[object] | object default: default value to be emitted to
                                              cli if this param is not available.
                                              if set to None, this param will not
                                              be emitted at all.
                                              defaults to None if not provided.

        :param str separator: separator to be used between values of list type.
                              defaults to single space if not provided.

        :keyword bool validate_index: specifies that index of this argument
                                      must be validated. it could be helpful
                                      to set this to False when there are multiple
                                      arguments with the same index that will appear
                                      in different situations.
                                      defaults to True if not provided.

        :raises ArgumentParamNameIsRequiredError: argument param name is required error.

        :raises PositionalArgumentIndexError: positional argument index error.
        """

        super().__init__(param_name, index, default, **options)

        if separator is None:
            separator = ' '

        self._separator = separator

    def _get_representation(self, value, is_default=False):
        """
        gets the representation of current cli option based on given input.

        it could be None if the value should not be emitted to cli.

        :param list[object] | object value: value of method input that
                                            should be represented in cli.

        :param bool is_default: specifies that input value is the default
                                value of this argument. it might be needed
                                for some subclasses to differentiate between
                                default value and other values.
                                defaults to False if not provided.

        :rtype: str
        """

        if value is None:
            return None

        result = value
        if isinstance(value, LIST_TYPES):
            result = self._separator.join([str(item) for item in value])

        return super()._get_representation(result, is_default)


class MappingArgument(KeywordArgumentBase):
    """
    mapping argument class.

    this class must be used for cli options that have mappings.
    these are arguments that their real value that should be emitted
    to cli, will be extracted from a dict based on the input of the
    python-side method. they could have argument name or could be emitted
    without any argument name.
    usually you don't need to use an instance of this class directly.
    """

    def __init__(self, param_name, param_value_to_cli_map,
                 cli_option_name=None, default=None):
        """
        initializes an instance of MappingArgument.

        :param str param_name: param name presented in method signature.

        :param dict param_value_to_cli_map: a dictionary containing a mapping
                                            between different method param values
                                            and their representation on cli. for
                                            example the `--autogenerate` flag of
                                            alembic, could be present to imply to
                                            `True` and absent to imply to `False`.
                                            so this dict should contain two keys,
                                            one is `True` and its value must be
                                            '--autogenerate' and another is `False`
                                            with the `None` value to prevent
                                            emitting to cli.

        :param str cli_option_name: relevant cli option name to `param_name`.
                                    it could be None if command does
                                    not have a name for this argument.
                                    for example the `--autogenerate` flag
                                    of alembic, could be present to imply to
                                    `True` and absent to imply to `False`.
                                    but `--message` flag of alembic must
                                    always be provided with given message.
                                    defaults to None if not provided.

        :param list[object] | object default: default value to be emitted to
                                              cli if this param is not available.
                                              if set to None, this param will not
                                              be emitted at all.
                                              defaults to None if not provided.

        :raises ArgumentParamNameIsRequiredError: argument param name is required error.

        :raises KeywordArgumentCLIOptionNameError: keyword argument cli option name error.

        :raises MappingArgumentParamValueToCLIMapRequiredError: mapping argument param value
                                                                to cli map required error.
        """

        super().__init__(param_name, cli_option_name, default)

        if param_value_to_cli_map is None or len(param_value_to_cli_map) <= 0:
            raise MappingArgumentParamValueToCLIMapRequiredError(
                'Mapping argument "param_value_to_cli_map" is '
                'required. it should contain at least one key/value.'
            )

        self._param_value_to_cli_map = param_value_to_cli_map

    def _get_representation(self, value, is_default=False):
        """
        gets the representation of current cli option based on given input.

        it could be None if the value should not be emitted to cli.

        :param list[object] | object value: value of method input that
                                            should be represented in cli.

        :param bool is_default: specifies that input value is the default
                                value of this argument. it might be needed
                                for some subclasses to differentiate between
                                default value and other values.
                                defaults to False if not provided.

        :rtype: list[object] | object
        """

        if is_default is False and value is not None:
            value = self._param_value_to_cli_map.get(value)

        return super()._get_representation(value, is_default)


class BooleanArgument(MappingArgument):
    """
    boolean argument class.

    this class must be used for boolean cli arguments.
    boolean arguments are those that will emit a keyword to imply to True
    and will not emit anything to imply to False. for example:
    `command --arg1` -> the value of `arg1` will be True in this form.
    `command` -> in this form the value of `arg1` will be False.
    """

    def __init__(self, param_name, true_value,
                 false_value=None, default=None):
        """
        initializes an instance of BooleanArgument.

        :param str param_name: param name presented in method signature.

        :param object true_value: the value that should be emitted to cli for True.

        :param object false_value: the value that should be emitted to cli
                                   for False. if not provided, defaults to
                                   None and will not be emitted.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.

        :raises ArgumentParamNameIsRequiredError: argument param name is required error.

        :raises BooleanArgumentValueError: boolean argument value error.
        """

        if true_value is None and false_value is None:
            raise BooleanArgumentValueError('Boolean argument "true_value" or '
                                            '"false_value" must be provided.')
        self._true_value = true_value
        self._false_value = false_value

        mapping = DTO()
        mapping[True] = true_value
        mapping[False] = false_value

        super().__init__(param_name, mapping, None, default)

    def _get_representation(self, value, is_default=False):
        """
        gets the representation of current cli option based on given input.

        it could be None if the value should not be emitted to cli.

        :param list[object] | object value: value of method input that
                                            should be represented in cli.

        :param bool is_default: specifies that input value is the default
                                value of this argument. it might be needed
                                for some subclasses to differentiate between
                                default value and other values.
                                defaults to False if not provided.

        :rtype: list[object] | object
        """

        if is_default is True and value is not None:
            value = value is True
            is_default = False

        return super()._get_representation(value, is_default)


class KeywordArgument(KeywordArgumentBase):
    """
    keyword argument class.

    it should be used for keyword argument cli options.
    keyword arguments are those that have to emit an argument name
    with an argument value, note that this type of arguments could not
    emit a list for a single argument. instead, if the value for a single
    argument is a list, each item in the list will be emitted separately
    with `--arg` name. for example:
    `command --arg1 value1` -> for single value.
    `command --arg1 item1 --arg1 item2 --arg1 item3` -> for a list value.
    if the cli handler supports keyword arguments with list values, you should
    use `CompositeKeywordArgument` class.
    """

    def __init__(self, param_name, cli_option_name, default=None):
        """
        initializes an instance of KeywordArgument.

        :param str param_name: param name presented in method signature.

        :param str cli_option_name: relevant cli option name to `param_name`.
                                    for example `--message` flag of alembic
                                    must be present with the message value itself.

        :param list[object] | object default: default value to be emitted to
                                              cli if this param is not available.
                                              if set to None, this param will not
                                              be emitted at all.
                                              defaults to None if not provided.

        :raises ArgumentParamNameIsRequiredError: argument param name is required error.

        :raises KeywordArgumentCLIOptionNameRequiredError: keyword argument cli option
                                                           name required error.
        """

        if cli_option_name in (None, '') or cli_option_name.isspace():
            raise KeywordArgumentCLIOptionNameRequiredError('Keyword argument '
                                                            '"cli_option_name" '
                                                            'must be provided.')

        super().__init__(param_name, cli_option_name, default)


class CompositeKeywordArgument(KeywordArgument):
    """
    composite keyword argument class.

    it should be used for composite keyword argument cli options.
    composite keyword arguments are those that have to emit an argument
    name with an argument value, note that this type of arguments could
    emit a list for a single argument with `--arg` name. for example:
    `command --arg1 value1` -> for single value.
    `command --arg1 item1 item2 item3` -> for a list value.
    if the cli handler does not support keyword arguments with list
    values, you should use `KeywordArgument` class.
    """

    def __init__(self, param_name, cli_option_name, separator=None, default=None):
        """
        initializes an instance of CompositeKeywordArgument.

        :param str param_name: param name presented in method signature.

        :param str cli_option_name: relevant cli option name to `param_name`.
                                    for example `--message` flag of alembic
                                    must be present with the message value itself.

        :param str separator: separator to be used between values of list type.
                              defaults to single space if not provided.

        :param list[object] | object default: default value to be emitted to
                                              cli if this param is not available.
                                              if set to None, this param will not
                                              be emitted at all.
                                              defaults to None if not provided.

        :raises ArgumentParamNameIsRequiredError: argument param name is required error.

        :raises KeywordArgumentCLIOptionNameRequiredError: keyword argument cli
                                                           option name required error.
        """

        super().__init__(param_name, cli_option_name, default)

        if separator is None:
            separator = ' '

        self._separator = separator

    def _merge_inputs(self, value):
        """
        merges the cli option name and given value to be emitted to cli.

        note that the values of list type, must be emitted using single
        `cli_option_name` separated with optional separator or space.
        for example if value = [1, 2, 3] and `cli_option_name` is `--num`
        and separator is `-` then it should be emitted to cli like this:
        ['--num', '1-2-3'] and if separator is None, then it should be
        emitted like this: ['--num', '1 2 3']
        if the cli handler does not support multiple values with single
        name (for example alembic commands), you should use
        `KeywordArgument` class instead.

        :param list[object] | object value: value to be emitted to cli.

        :rtype: list[object]
        """

        if value is None:
            return None

        result = [self._cli_option_name]
        if isinstance(value, LIST_TYPES):
            value = self._separator.join([str(item) for item in value])

        result.append(value)
        return result


class JSONKeywordArgument(KeywordArgument):
    """
    json keyword argument class.

    it should be used for json keyword argument cli options.
    json keyword arguments are those that have to emit an argument
    name with an argument value, note that this type of arguments must
    be emitted as a list or dict json string representation. for example:
    `command --arg1 '[value1, value2]'` -> for list value.
    `command --arg1 '{"key1": item1, "key2": item2}'` -> for a dict value.
    """

    def _merge_inputs(self, value):
        """
        merges the cli option name and given value to be emitted to cli.

        :param list[object] | dict[object] value: value to be emitted to cli.

        :rtype: list[object]
        """

        if value is None:
            return None

        if isinstance(value, (tuple, set)):
            value = list(value)

        result = [self._cli_option_name]
        value = flask_json.dumps(value)
        result.append(value)
        return result



# -*- coding: utf-8 -*-
"""
database decorators module.
"""

import pyrin.database.services as database_services


def session_factory(*args, **kwargs):
    """
    decorator to register a database session factory.

    :param object args: session factory class constructor arguments.
    :param object kwargs: session factory class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           session factory with the same name, replace it with
                           the new one, otherwise raise an error. defaults to False.

    :raises InvalidSessionFactoryTypeError: invalid session factory type error.
    :raises DuplicatedSessionFactoryError: duplicated session factory error.

    :returns: session factory class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into database session factories.

        :param type cls: session factory class.

        :returns: session factory class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        database_services.register_session_factory(instance, **kwargs)

        return cls

    return decorator


def database_hook():
    """
    decorator to register a database hook.

    :raises InvalidDatabaseHookTypeError: invalid database hook type error.

    :returns: database hook class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available database hooks.

        :param type cls: database hook class.

        :returns: database hook class.
        :rtype: type
        """

        instance = cls()
        database_services.register_hook(instance)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
celery cli handlers worker module.
"""

from pyrin.task_queues.celery.cli.decorators import celery_cli_handler
from pyrin.task_queues.celery.cli.enumerations import CeleryCLIHandlersEnum
from pyrin.task_queues.celery.cli.interface import CeleryCLIHandlerBase
from pyrin.task_queues.celery.cli.handlers.params import ConcurrencyParam, \
    PurgeParam, BeatParam, HostnameParam, AutoScaleParam, WorkerLogFileParam, \
    WorkerPIDFileParam, WorkerLogLevelParam, OptimizationParam, WorkerQueuesParam


@celery_cli_handler()
class WorkerCLIHandler(CeleryCLIHandlerBase):
    """
    worker cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of WorkerCLIHandler.
        """

        super().__init__(CeleryCLIHandlersEnum.WORKER)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([ConcurrencyParam(), PurgeParam(), BeatParam(),
                       HostnameParam(), AutoScaleParam(), WorkerLogFileParam(),
                       WorkerPIDFileParam(), WorkerLogLevelParam(), WorkerQueuesParam(),
                       OptimizationParam()])

        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
cors structs module.
"""

import pyrin.application.services as application_services

from pyrin.core.structs import CoreObject
from pyrin.processor.cors.exceptions import CORSAllowedHeadersModificationError


class CORS(CoreObject):
    """
    cors class.
    """

    def __init__(self, **options):
        """
        initializes an instance of `CORS`.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    if not provided, it will be get from cors config store.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.
        """

        super().__init__()

        self._enabled = options.get('cors_enabled')
        self._always_send = options.get('cors_always_send')
        self._allowed_origins = options.get('cors_allowed_origins')
        self._exposed_headers = options.get('cors_exposed_headers')
        self._allowed_headers = options.get('cors_allowed_headers')
        self._allow_credentials = options.get('cors_allow_credentials')
        self._max_age = options.get('cors_max_age')

    def add_allowed_headers(self, *headers):
        """
        adds given headers into allowed headers.

        :param str headers: headers to be added.

        :raises CORSAllowedHeadersModificationError: cors allowed headers modification error.
        """

        if application_services.got_first_request() is True:
            raise CORSAllowedHeadersModificationError('Cors allowed headers could not be '
                                                      'modified after first request is served.')

        if headers is not None and len(headers) > 0:
            current_headers = self._allowed_headers or []
            current_headers = set(current_headers).union(set(headers))
            self._allowed_headers = list(current_headers)

    @property
    def enabled(self):
        """
        gets a value indicating that cross origin resource sharing is enabled.

        :rtype: bool
        """

        return self._enabled

    @property
    def always_send(self):
        """
        gets a value indicating that cors headers must be included in response.

        even if the request does not have origin header.

        :rtype: bool
        """

        return self._always_send

    @property
    def allowed_origins(self):
        """
        gets a list of extra allowed origins to be used in conjunction with default allowed ones.

        :rtype: list[str]
        """

        return self._allowed_origins

    @property
    def allowed_headers(self):
        """
        gets a list of extra allowed headers to be used in conjunction with default allowed ones.

        :rtype: list[str]
        """

        return self._allowed_headers

    @property
    def exposed_headers(self):
        """
        gets a list of extra exposed headers to be used in conjunction with default exposed ones.

        :rtype: list[str]
        """

        return self._exposed_headers

    @property
    def allow_credentials(self):
        """
        gets a value indicating that browsers are allowed to pass response headers.

        to front-end javascript code.

        :rtype: bool
        """

        return self._allow_credentials

    @property
    def max_age(self):
        """
        gets the maximum number of seconds to cache results.

        :rtype: int
        """

        return self._max_age



# -*- coding: utf-8 -*-
"""
cli core exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class CLICoreManagerException(CoreException):
    """
    cli core manager exception.
    """
    pass


class CLICoreManagerBusinessException(CoreBusinessException, CLICoreManagerException):
    """
    cli core manager business exception.
    """
    pass


class CLICoreTemplateHandlerNotFoundError(CLICoreManagerBusinessException):
    """
    cli core template handler not found error.
    """
    pass



# -*- coding: utf-8 -*-
"""
orm sql extractor component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.database.orm.sql.extractor import ORMSQLExtractorPackage
from pyrin.database.orm.sql.extractor.manager import ORMSQLExtractorManager


@component(ORMSQLExtractorPackage.COMPONENT_NAME)
class ORMSQLExtractorComponent(Component, ORMSQLExtractorManager):
    """
    orm sql extractor component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
core globals module.
"""

from flask_babel import gettext
from sqlalchemy.engine import Row


class Constant:
    """
    constant class.

    this is a helper class for defining global constant objects.
    we have to declare this class in `globals` instead of `structs`
    to prevent circular dependency.
    """

    def __init__(self, name, **options):
        """
        initializes an instance of Constant.

        :param str name: this constant's name.
        """

        super().__init__()
        self._name = name

    def __repr__(self):
        """
        gets the string representation of current constant.

        :rtype: str
        """

        return str(self)

    def __str__(self):
        """
        gets the string representation of current constant.

        :rtype: str
        """

        return self._name


# this value should be used as `None`, where the `None` itself has a meaning.
NULL = Constant('NULL')

# this value should be used where we need to reference class type of None objects.
NONE_TYPE = type(None)

# this value should be used where we want to determine if an object is a list type or not.
LIST_TYPES = (list, tuple, set)

# this method should be used for localizable strings.
_ = gettext

# this value should be used when working on Row objects.
ROW_RESULT = Row

# this value should be used instead of True when we want to preserve security of
# inputs which come from client to prevent injection of invalid values through options.
SECURE_TRUE = Constant('SECURE_TRUE')

# this value should be used instead of False when we want to preserve security of
# inputs which come from client to prevent injection of invalid values through options.
SECURE_FALSE = Constant('SECURE_FALSE')



# -*- coding: utf-8 -*-
"""
orm sql operators base module.
"""

from datetime import datetime

from sqlalchemy.sql import Select
from sqlalchemy.sql.elements import BindParameter
from sqlalchemy.sql.operators import ColumnOperators

import pyrin.utils.datetime as datetime_utils
import pyrin.utils.misc as misc_utils

from pyrin.utils.sqlalchemy import like_prefix, like_exact_prefix, \
    like_suffix, like_exact_suffix


class CoreColumnOperators(ColumnOperators):
    """
    core column operators class.

    this class provides some practical benefits
    to its subclasses for sql operations.
    """

    DEFAULT_ESCAPE_CHAR = '/'

    def _process_like_autoescape(self, value, escape, autoescape):
        """
        processes the value based on autoescape flag for like command.

        it may return an escaped string or the exact input string.

        :param str value: value to be escaped.

        :param str escape: a character to be used as escape character.
                           defaults to `DEFAULT_ESCAPE_CHAR` if not provided.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :rtype: str
        """

        if autoescape is True:
            if escape is None:
                escape = self.DEFAULT_ESCAPE_CHAR

            if escape not in ("%", "_"):
                value = value.replace(escape, escape + escape)

            value = value.replace("%", escape + "%").replace("_", escape + "_")

        return value

    def _process_like_prefix(self, value, **options):
        """
        processes the value that should be prefixed to

        value for `like` expression based on given options.

        :param str value: expression to be compared.

        :keyword bool exact_start: specifies that value should be emitted
                                   without any modifications at start of it.
                                   defaults to True if not provided.

        :keyword int start_count: count of `_` chars to be attached to beginning.
                                  if not provided, `%` will be used.

        :note start_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                           if the provided value goes upper than this limit,
                           a `%` will be attached instead of it. this limit
                           is for security reason.

        :rtype: str
        """

        exact_start = options.get('exact_start', True)
        if exact_start is False:
            begin_wrapper = like_prefix
            inputs = (value,)
            begin_count = options.get('start_count', None)
            if begin_count is not None:
                begin_wrapper = like_exact_prefix
                inputs = (value, begin_count)

            value = begin_wrapper(*inputs)

        return value

    def _process_like_suffix(self, value, **options):
        """
        processes the value that should be suffixed to
        value for `like` expression based on given options.

        :param str value: expression to be compared.

        :keyword bool exact_end: specifies that value should be emitted
                                 without any modifications at end of it.
                                 defaults to True if not provided.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                         if the provided value goes upper than this limit,
                         a `%` will be attached instead of it. this limit
                         is for security reason.

        :rtype: str
        """

        exact_end = options.get('exact_end', True)
        if exact_end is False:
            end_wrapper = like_suffix
            inputs = (value,)
            end_count = options.get('end_count', None)
            if end_count is not None:
                end_wrapper = like_exact_suffix
                inputs = (value, end_count)

            value = end_wrapper(*inputs)

        return value

    def between_datetime(self, cleft, cright, symmetric=False, **options):
        """
        produces an `expression.between` clause against the parent object.

        given the lower and upper datetime range. this method is
        implemented to be able to handle datetime values more practical.

        if the inputs are date objects, they will be converted to datetime with client
        timezone and `consider_begin_of_day` and `consider_end_of_day` will also
        considered as True.

        :param datetime | date cleft: lower bound of datetime clause.
        :param datetime | date cright: upper bound of datetime clause.

        :param bool symmetric: specifies to emmit `between symmetric` to database.
                               note that not all databases support symmetric.
                               but `between symmetric` is equivalent to
                               `between least(a, b) and greatest(a, b)`.

        :keyword bool consider_begin_of_day: specifies that consider begin
                                             of day for lower datetime.
                                             defaults to False if not provided.

        :keyword bool consider_end_of_day: specifies that consider end
                                           of day for upper datetime.
                                           defaults to False if not provided.
        """

        cleft, cright = datetime_utils.normalize_datetime_range(cleft, cright, **options)
        return self.between(cleft, cright, symmetric=symmetric)

    def istartswith(self, other, **options):
        """
        implements the `startswith` operator.

        produces an `ilike` expression that tests against a match for the
        start of a string value. for example:
        column like <other> || `%` or column like <other> || `_`
        this method provides a case-insensitive variant of `startswith()` method.

        :param str other: expression to be compared.

        :keyword str escape: optional escape character,
                             renders the `escape` keyword.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                         if the provided value goes upper than this limit,
                         a `%` will be attached instead of it. this limit
                         is for security reason.
        """

        escape = options.pop('escape', None)
        options.update(exact_end=False)
        other = self._process_like_autoescape(other, escape, options.get('autoescape', False))
        other = self._process_like_suffix(other, **options)
        return self.ilike(other, escape)

    def iendswith(self, other, **options):
        """
        implements the `endswith` operator.

        produces an `ilike` expression that tests against a match for the
        end of a string value. for example:
        column like '%' || <other> or column like '_' || <other>
        this method provides a case-insensitive variant of `endswith()` method.

        :param str other: expression to be compared.

        :keyword str escape: optional escape character,
                             renders the `escape` keyword.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :keyword int start_count: count of `_` chars to be attached to start.
                                  if not provided, `%` will be used.

        :note start_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                           if the provided value goes upper than this limit,
                           a `%` will be attached instead of it. this limit
                           is for security reason.
        """

        escape = options.pop('escape', None)
        options.update(exact_start=False)
        other = self._process_like_autoescape(other, escape, options.get('autoescape', False))
        other = self._process_like_prefix(other, **options)
        return self.ilike(other, escape)

    def icontains(self, other, **options):
        """
        implements the 'contains' operator.

        produces an `ilike` expression that tests against a match for the
        middle of a string value. for example:
        column like `%` || <other> || `%` or column like `_` || <other> || `_`
        this method provides a case-insensitive variant of `contains()` method.

        :param str other: expression to be compared.

        :keyword str escape: optional escape character,
                             renders the `escape` keyword.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :keyword int start_count: count of `_` chars to be attached to beginning.
                                  if not provided, `%` will be used.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note start_count, end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                                      if the provided value goes upper than this limit,
                                      a `%` will be attached instead of it. this limit
                                      is for security reason.
        """

        escape = options.pop('escape', None)
        options.update(exact_start=False, exact_end=False)
        other = self._process_like_autoescape(other, escape, options.get('autoescape', False))
        other = self._process_like_prefix(other, **options)
        other = self._process_like_suffix(other, **options)
        return self.ilike(other, escape)

    def in_(self, other):
        """
        implements the `in` operator.

        in a column context, produces the clause `column in other`.

        the given parameter `other` may be:

        * a list of literal values.
        * a list of tuples may be provided if the comparison is against a
          `.tuple_` containing multiple expressions.
        * an empty list.
        * a bound parameter, for example `.bindparam`, may be used if it
          includes the `.bindparam.expanding` flag.
        * an `_expression.select` construct, which is usually a correlated
          scalar select.
        * a single literal value could also be set and it will be converted into iterable.

        :param list | object | BindParameter | Select other: a list of literals, a
                                                             `_expression.select` construct,
                                                             or a `.bindparam` construct that
                                                             includes the `.bindparam.expanding`
                                                             flag set to True.
        """

        if not isinstance(other, (list, tuple, set, Select, BindParameter)):
            other = misc_utils.make_iterable(other, list)

        return super().in_(other)



# -*- coding: utf-8 -*-
"""
celery audit component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.task_queues.celery.audit import CeleryAuditPackage
from pyrin.task_queues.celery.audit.manager import CeleryAuditManager


@component(CeleryAuditPackage.COMPONENT_NAME)
class CeleryAuditComponent(Component, CeleryAuditManager):
    """
    celery audit component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
validator auto package.
"""

from pyrin.packaging.base import Package


class ValidatorAutoPackage(Package):
    """
    validator auto package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'validator.auto.component'
    DEPENDS = ['pyrin.database.model']
